#!/usr/bin/env python
"""Sample client for using pyadtpulse."""

import logging
import argparse
import asyncio
import json
import sys
from pprint import pprint
from time import sleep

from pyadtpulse import PyADTPulse
from pyadtpulse.const import (
    ADT_DEFAULT_KEEPALIVE_INTERVAL,
    ADT_DEFAULT_POLL_INTERVAL,
    ADT_DEFAULT_RELOGIN_INTERVAL,
    API_HOST_CA,
    DEFAULT_API_HOST,
)
from pyadtpulse.site import ADTPulseSite
from pyadtpulse.util import AuthenticationException

USER = "adtpulse_user"
PASSWD = "adtpulse_password"
FINGERPRINT = "adtpulse_fingerprint"
PULSE_DEBUG = "debug"
TEST_ALARM = "test_alarm"
SLEEP_INTERVAL = "sleep_interval"
USE_ASYNC = "use_async"
DEBUG_LOCKS = "debug_locks"
KEEPALIVE_INTERVAL = "keepalive_interval"
RELOGIN_INTERVAL = "relogin_interval"
SERVICE_HOST = "service_host"
POLL_INTERVAL = "poll_interval"
DETAILED_DEBUG_LOGGING = "detailed_debug_logging"

BOOLEAN_PARAMS = {
    USE_ASYNC,
    DEBUG_LOCKS,
    PULSE_DEBUG,
    TEST_ALARM,
    DETAILED_DEBUG_LOGGING,
}
INT_PARAMS = {SLEEP_INTERVAL, KEEPALIVE_INTERVAL, RELOGIN_INTERVAL}
FLOAT_PARAMS = {POLL_INTERVAL}

# Default values
DEFAULT_USE_ASYNC = True
DEFAULT_DEBUG = False
DEFAULT_DETAILED_DEBUG_LOGGING = False
DEFAULT_TEST_ALARM = False
DEFAULT_SLEEP_INTERVAL = 5
DEFAULT_DEBUG_LOCKS = False


# Constants for environment variable names
ENV_USER = "USER"
ENV_PASSWORD = "PASSWORD"
ENV_FINGERPRINT = "FINGERPRINT"


def setup_logger(level: int):
    """Set up logger."""
    logger = logging.getLogger()
    logger.setLevel(level)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(level)

    formatter = logging.Formatter("%(asctime)s %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def handle_args() -> argparse.Namespace:
    """Handle program arguments using argparse.

    Returns:
        argparse.Namespace: Parsed command-line arguments.
    """
    parser = argparse.ArgumentParser(description="ADT Pulse example client")
    parser.add_argument("json_file", nargs="?", help="JSON file containing parameters")
    parser.add_argument(
        f"--{USER}",
        help="Pulse username (can be set in JSON file or environment variable)",
    )
    parser.add_argument(
        f"--{PASSWD}",
        help="Pulse password (can be set in JSON file or environment variable)",
    )
    parser.add_argument(
        f"--{FINGERPRINT}",
        help="Pulse fingerprint (can be set in JSON file or environment variable)",
    )
    parser.add_argument(
        f"--{SERVICE_HOST}",
        help=f"Pulse service host, must be {DEFAULT_API_HOST} or {API_HOST_CA}, "
        f"default is {DEFAULT_API_HOST}",
    )
    parser.add_argument(
        f"--{PULSE_DEBUG}",
        type=bool,
        default=None,
        help="Set True to enable debugging",
    )
    parser.add_argument(
        f"--{DETAILED_DEBUG_LOGGING}",
        type=bool,
        default=None,
        help="Set True to enable detailed debug logging",
    )
    parser.add_argument(
        f"--{TEST_ALARM}",
        type=bool,
        default=None,
        help="Set True to test alarm arming/disarming",
    )
    parser.add_argument(
        f"--{SLEEP_INTERVAL}",
        type=int,
        default=None,
        help="Number of seconds to sleep between each call "
        f"(default: {DEFAULT_SLEEP_INTERVAL} seconds),"
        " not used for async",
    )
    parser.add_argument(
        f"--{USE_ASYNC}",
        type=bool,
        default=None,
        help=f"Set to true to use asyncio (default: {DEFAULT_USE_ASYNC})",
    )
    parser.add_argument(
        f"--{DEBUG_LOCKS}",
        type=bool,
        default=None,
        help=f"Set to true to debug thread locks, default: {DEFAULT_DEBUG_LOCKS}",
    )
    parser.add_argument(
        f"--{KEEPALIVE_INTERVAL}",
        type=int,
        default=None,
        help="Number of minutes to wait between keepalive calls (default: "
        f"{ADT_DEFAULT_KEEPALIVE_INTERVAL} minutes)",
    )
    parser.add_argument(
        f"--{RELOGIN_INTERVAL}",
        type=int,
        default=None,
        help="Number of minutes to wait between relogin calls "
        f"(default: {ADT_DEFAULT_RELOGIN_INTERVAL} minutes)",
    )

    parser.add_argument(
        f"--{POLL_INTERVAL}",
        type=float,
        default=None,
        help="Number of seconds to wait between polling calls "
        f"(default: {ADT_DEFAULT_POLL_INTERVAL} seconds)",
    )

    args = parser.parse_args()

    json_params = load_parameters_from_json(args.json_file)

    # Update arguments with values from the JSON file
    # load_parameters_from_json() will handle incorrect types
    if json_params is not None:
        for key, value in json_params.items():
            if getattr(args, key) is None and value is not None:
                setattr(args, key, value)

    # Set default values for specific parameters
    args.use_async = args.use_async if args.use_async is not None else DEFAULT_USE_ASYNC
    args.debug_locks = (
        args.debug_locks if args.debug_locks is not None else DEFAULT_DEBUG_LOCKS
    )
    args.debug = args.debug if args.debug is not None else DEFAULT_DEBUG
    args.detailed_debug_logging = (
        args.detailed_debug_logging
        if args.detailed_debug_logging is not None
        else DEFAULT_DETAILED_DEBUG_LOGGING
    )
    args.test_alarm = (
        args.test_alarm if args.test_alarm is not None else DEFAULT_TEST_ALARM
    )
    if args.use_async is False and args.sleep_interval is None:
        args.sleep_interval = DEFAULT_SLEEP_INTERVAL
    args.keepalive_interval = (
        args.keepalive_interval
        if args.keepalive_interval is not None
        else ADT_DEFAULT_KEEPALIVE_INTERVAL
    )
    args.relogin_interval = (
        args.relogin_interval
        if args.relogin_interval is not None
        else ADT_DEFAULT_RELOGIN_INTERVAL
    )
    args.service_host = (
        args.service_host if args.service_host is not None else DEFAULT_API_HOST
    )
    args.poll_interval = (
        args.poll_interval
        if args.poll_interval is not None
        else ADT_DEFAULT_POLL_INTERVAL
    )

    return args


def load_parameters_from_json(json_file: str) -> dict | None:
    """Load parameters from a JSON file.

    Args:
        json_file (str): Path to the JSON file.

    Returns:
        Optional[Dict]: Loaded parameters as a dictionary,
                        or None if there was an error.
    """
    try:
        with open(json_file, encoding="utf-8") as file:
            parameters = json.load(file)
            invalid_keys = []
            for key, value in parameters.items():
                if key in BOOLEAN_PARAMS and not isinstance(value, bool):
                    print(
                        "Invalid boolean value for "
                        f"{key}: {value}"
                        " in JSON file, ignoring..."
                    )
                    invalid_keys.append(key)
                elif key in INT_PARAMS and not isinstance(value, int):
                    print(
                        "Invalid integer value for "
                        f"{key}: {value}"
                        " in JSON file, ignoring..."
                    )
                    invalid_keys.append(key)
                elif (
                    key in FLOAT_PARAMS
                    and not isinstance(value, float)
                    and not isinstance(value, int)
                ):
                    print(
                        "Invalid float value for "
                        f"{key}: {value}"
                        " in JSON file, ignoring..."
                    )
                    invalid_keys.append(key)
            for key in invalid_keys:
                del parameters[key]
            return parameters
    except FileNotFoundError:
        print(f"JSON file not found: {json_file}")
        return None
    except json.JSONDecodeError as e:
        print(f"Error parsing JSON: {e}")
        return None


def print_site(site: ADTPulseSite) -> None:
    """Print site information.

    Args:
        site (ADTPulseSite): The site to display
    """
    print(f"Site: {site.name} (id={site.id})")
    print("Alarm panel: ")
    pprint(site.alarm_control_panel, compact=True)
    print("Gateway: ")
    pprint(site.gateway, compact=True)


def check_updates(site: ADTPulseSite, adt: PyADTPulse, run_alarm_test: bool) -> bool:
    """Check a site for updates and print details.

    We don't really need to do this anymore, as the data will be
    updated in the background

    Args:
        site (ADTPulseSite): site to check
        adt (PyADTPulse): Pulse connection object
        test_alarm: bool: sleep a bit if testing alarm

        Returns: bool: True if update successful
    """
    if run_alarm_test:
        while (
            site.alarm_control_panel.is_arming or site.alarm_control_panel.is_disarming
        ):
            print(
                f"site is_arming: {site.alarm_control_panel.is_arming}, "
                f"site is_disarming: {site.alarm_control_panel.is_disarming}"
            )
            sleep(1)
    print(f"Gateway online: {adt.site.gateway.is_online}")
    if adt.update():
        print("ADT Data updated, at " f"{site.last_updated}, refreshing")
        return True
    print("Site update failed")
    return False


def test_alarm(site: ADTPulseSite, adt: PyADTPulse) -> None:
    """Test alarm functions.

    Args:
        site (ADTPulseSite): site to test
        adt (PyADTPulse): ADT Pulse connection objecct
        sleep_interval (int): length to sleep between tests
    """
    print("Arming alarm stay")
    if site.arm_home():
        print("Alarm arming home succeeded")
        check_updates(site, adt, True)
        print("Testing invalid alarm state change from armed home to armed away")
        if site.arm_away():
            print("Error, armed away while already armed")
        else:
            print("Test succeeded")
            print("Testing changing alarm status to same value")
            if site.arm_home():
                print("Error, allowed arming to same state")
            else:
                print("Test succeeded")
    else:
        print("Alarm arming home failed, attempting force arm")
        if site.arm_home(True):
            print("Force arm succeeded")
        else:
            print("Force arm failed")

    print()
    print_site(site)

    print("Disarming alarm")
    if site.disarm():
        print("Disarming succeeded")
        check_updates(site, adt, True)
        print("Testing disarming twice")
        if site.disarm():
            print("Test disarm twice suceeded")
        else:
            print("Test disarm twice failed")
    else:
        print("Disarming failed")

    print()
    print_site(site)
    print("Arming alarm away")

    if site.arm_away():
        print("Arm away succeeded")
        check_updates(site, adt, True)
    else:
        print("Arm away failed")

    print()
    print_site(site)
    site.disarm()
    print("Disarmed")


def sync_example(
    username: str,
    password: str,
    fingerprint: str,
    run_alarm_test: bool,
    sleep_interval: int,
    debug_locks: bool,
    poll_interval: float,
    keepalive_interval: int,
    relogin_interval: int,
    detailed_debug_logging: bool,
) -> None:
    """Run example of sync pyadtpulse calls.

    Args:
        username (str): Pulse username
        password (str): Pulse password
        fingerprint (str): Pulse fingerprint
        run_alarm_test (bool): True if alarm test to be run
        sleep_interval (int): how long in seconds to sleep between update checks
        debug_locks: bool: True to enable thread lock debugging
        keepalive_interval (int): keepalive interval in minutes
        relogin_interval (int): relogin interval in minutes
        detailed_debug_logging (bool): True to enable detailed debug logging
    """
    try:
        adt = PyADTPulse(
            username,
            password,
            fingerprint,
            debug_locks=debug_locks,
            keepalive_interval=keepalive_interval,
            relogin_interval=relogin_interval,
            detailed_debug_logging=detailed_debug_logging,
        )
    except AuthenticationException:
        print("Invalid credentials for ADT Pulse site")
        sys.exit()
    except BaseException as e:
        print("Received exception logging into ADT Pulse site")
        print(f"{e}")
        sys.exit()

    if not adt.is_connected:
        print("Error: Could not log into ADT Pulse site")
        return
    if len(adt.sites) == 0:
        print("Error: could not retrieve sites")
        adt.logout()
        return

    adt.site.site_lock.acquire()
    print(f"Gateway online: {adt.site.gateway.is_online}")
    print_site(adt.site)
    if not adt.site.zones:
        print("Error: no zones exist, exiting")
        adt.site.site_lock.release()
        adt.logout()
        return
    adt.site.gateway.poll_interval = poll_interval
    pprint(adt.site.zones, compact=True)
    adt.site.site_lock.release()
    if run_alarm_test:
        test_alarm(adt.site, adt)

    done = False
    while not done:
        try:
            print_site(adt.site)
            print("----")
            if not adt.site.zones:
                print("Error, no zones exist, exiting...")
                done = True
                break
            if adt.updates_exist:
                print("Updates exist, refreshing")
                # Don't need to explicitly call update() anymore
                # Background thread will already have updated
                if not adt.update():
                    print("Error occurred fetching updates, exiting..")
                    done = True
                    break
                print("\nZones:")
                with adt.site.site_lock:
                    pprint(adt.site.zones, compact=True)
            else:
                print("No updates exist")
            sleep(sleep_interval)
        except KeyboardInterrupt:
            print("exiting...")
            done = True

    print("Logging out")
    adt.logout()


async def async_test_alarm(adt: PyADTPulse) -> None:
    """Test alarm functions.

    Args:
        site (ADTPulseSite): site to test
        adt (PyADTPulse): ADT Pulse connection objecct
    """
    print("Arming alarm stay")
    if await adt.site.async_arm_home():
        print("Arming stay call succeeded")
        if adt.site.alarm_control_panel.is_arming:
            print("Arming stay pending check succeeded")
        else:
            print(
                "FAIL: Arming home pending check failed "
                f"{adt.site.alarm_control_panel} "
            )
        if await adt.wait_for_update():
            if adt.site.alarm_control_panel.is_home:
                print("Arm stay no longer pending")
            else:
                while not adt.site.alarm_control_panel.is_home:
                    pprint(
                        f"FAIL: Arm stay value incorrect {adt.site.alarm_control_panel}"
                    )
                    if not await adt.wait_for_update():
                        print("ERROR: Alarm update failed")
                        break
        else:
            print("ERROR: Alarm update failed")

        print("Testing invalid alarm state change from armed home to armed away")
        if await adt.site.async_arm_away():
            print(
                f"FAIL: armed away while already armed {adt.site.alarm_control_panel}"
            )
        else:
            print("Test succeeded")
            print("Testing changing alarm status to same value")
            if await adt.site.async_arm_home():
                print(
                    f"FAIL: allowed arming to same state {adt.site.alarm_control_panel}"
                )
            else:
                print("Test succeeded")

    else:
        print("Alarm arming home failed, attempting force arm")
        if await adt.site.async_arm_home(True):
            print("Force arm succeeded")
        else:
            print(f"FAIL: Force arm failed {adt.site.alarm_control_panel}")
    print("Disarming alarm")
    if await adt.site.async_disarm():
        print("Disarming succeeded")
        if adt.site.alarm_control_panel.is_disarming:
            print("Disarm pending success")
        else:
            pprint(f"FAIL: Disarm pending fail {adt.site.alarm_control_panel}")
        if await adt.wait_for_update():
            if adt.site.alarm_control_panel.is_disarmed:
                print("Success update to disarm")
            else:
                while not adt.site.alarm_control_panel.is_disarmed:
                    pprint(
                        "FAIL: did not set to disarm after update "
                        f"{adt.site.alarm_control_panel}"
                    )
                    if not await adt.wait_for_update():
                        print("ERROR: Alarm update failed")
                        break
            print("Test finally succeeded")
        print("Testing disarming twice")
        if await adt.site.async_disarm():
            print("Double disarm call succeeded")
        else:
            pprint(f"FAIL: Double disarm call failed {adt.site.alarm_control_panel}")
        if adt.site.alarm_control_panel.is_disarming:
            print("Double disarm state is disarming")
        else:
            pprint(
                "FAIL: Double disarm state is not disarming "
                f"{adt.site.alarm_control_panel}"
            )
        if await adt.wait_for_update():
            if adt.site.alarm_control_panel.is_disarmed:
                print("Double disarm success")
            else:
                while not adt.site.alarm_control_panel.is_disarmed:
                    pprint(
                        "FAIL: Double disarm state is not disarmed "
                        f"{adt.site.alarm_control_panel}"
                    )
                    if not await adt.wait_for_update():
                        print("ERROR: Alarm update failed")
                        break
            print("Test finally succeeded")
        else:
            print("ERROR: Alarm update failed")
    else:
        print("Disarming failed")
    print("Arming alarm away")
    if await adt.site.async_arm_away():
        print("Arm away call succeeded")
        if adt.site.alarm_control_panel.is_arming:
            print("Arm away arm pending")
        else:
            pprint(f"FAIL: arm away call not pending {adt.site.alarm_control_panel}")
        if await adt.wait_for_update():
            if adt.site.alarm_control_panel.is_away:
                print("Arm away call after update succeed")
            else:
                while not adt.site.alarm_control_panel.is_away:
                    pprint(
                        "FAIL: arm away call after update failed "
                        "f{adt.site.alarm_control_panel}"
                    )
                    if not await adt.wait_for_update():
                        print("ERROR: Alarm update failed")
                        break
            print("Test finally succeeded")
        else:
            print("ERROR: Alarm update failed")
    else:
        print("Arm away failed")
    await adt.site.async_disarm()
    print("Disarmed")


async def async_example(
    username: str,
    password: str,
    fingerprint: str,
    run_alarm_test: bool,
    debug_locks: bool,
    poll_interval: float,
    keepalive_interval: int,
    relogin_interval: int,
    detailed_debug_logging: bool,
) -> None:
    """Run example of pytadtpulse async usage.

    Args:
        username (str): Pulse username
        password (str): Pulse password
        fingerprint (str): Pulse fingerprint
        run_alarm_test (bool): True if alarm tests should be run
        debug_locks (bool): True to enable thread lock debugging
        poll_interval (float): polling interval in seconds
        keepalive_interval (int): keepalive interval in minutes
        relogin_interval (int): relogin interval in minutes
        detailed_debug_logging (bool): enable detailed debug logging
    """
    adt = PyADTPulse(
        username,
        password,
        fingerprint,
        do_login=False,
        debug_locks=debug_locks,
        keepalive_interval=keepalive_interval,
        relogin_interval=relogin_interval,
        detailed_debug_logging=detailed_debug_logging,
    )

    if not await adt.async_login():
        print("ADT Pulse login failed")
        return

    if not adt.is_connected:
        print("Error: could not log into ADT Pulse site")
        return

    if adt.site is None:
        print("Error: could not retrieve sites")
        await adt.async_logout()
        return

    print_site(adt.site)
    if adt.site.zones is None:
        print("Error: no zones exist")
        await adt.async_logout()
        return
    adt.site.gateway.poll_interval = poll_interval
    pprint(adt.site.zones, compact=True)
    if run_alarm_test:
        await async_test_alarm(adt)

    done = False
    while not done:
        try:
            print(f"Gateway online: {adt.site.gateway.is_online}")
            print_site(adt.site)
            print("----")
            if not adt.site.zones:
                print("No zones exist, exiting...")
                done = True
                break
            print("\nZones:")
            pprint(adt.site.zones, compact=True)
            if await adt.wait_for_update():
                print("Updates exist, refreshing")
                # no need to call an update method
            else:
                print("Warning, update failed, retrying")
        except KeyboardInterrupt:
            print("exiting...")
            done = True

    print("Logging out")
    await adt.async_logout()


def main():
    """Run main program."""
    args = handle_args()

    if not args or not any(
        [args.adtpulse_user, args.adtpulse_password, args.adtpulse_fingerprint]
    ):
        print(f"ERROR! {USER}, {PASSWD}, and {FINGERPRINT} must all be set")
        raise SystemExit

    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.ERROR

    use_async = args.use_async

    setup_logger(level)

    if not use_async:
        sync_example(
            args.adtpulse_user,
            args.adtpulse_password,
            args.adtpulse_fingerprint,
            args.test_alarm,
            args.sleep_interval,
            args.debug_locks,
            args.poll_interval,
            args.keepalive_interval,
            args.relogin_interval,
            args.detailed_debug_logging,
        )
    else:
        asyncio.run(
            async_example(
                args.adtpulse_user,
                args.adtpulse_password,
                args.adtpulse_fingerprint,
                args.test_alarm,
                args.debug_locks,
                args.poll_interval,
                args.keepalive_interval,
                args.relogin_interval,
                args.detailed_debug_logging,
            )
        )


if __name__ == "__main__":
    main()



import logging

from pyadtpulse.util import remove_prefix

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.DEBUG)


class TestRemovePrefix:
    # prefix is at the beginning of the text
    def test_prefix_at_beginning(self):
        assert remove_prefix("hello world", "hello") == " world"

    # prefix is not in the text
    def test_prefix_not_in_text(self):
        assert remove_prefix("hello world", "hi") == "hello world"

    # prefix is an empty string
    def test_empty_prefix(self):
        assert remove_prefix("hello world", "") == "hello world"

    # prefix is the entire text
    def test_entire_text_as_prefix(self):
        assert remove_prefix("hello world", "hello world") == ""

    # prefix is longer than the text
    def test_longer_prefix(self):
        assert remove_prefix("hello", "hello world") == "hello"

    # text is an empty string
    def test_empty_text(self):
        assert remove_prefix("", "hello") == ""



"""Constants for pyadtpulse."""
__version__ = "1.1.4b2"

from enum import Enum
from http import HTTPStatus

DEFAULT_API_HOST = "https://portal.adtpulse.com"
API_HOST_CA = "https://portal-ca.adtpulse.com"  # Canada

API_PREFIX = "/myhome/"

ADT_LOGIN_URI = "/access/signin.jsp"
ADT_LOGOUT_URI = "/access/signout.jsp"

ADT_SUMMARY_URI = "/summary/summary.jsp"
ADT_ZONES_URI = "/ajax/homeViewDevAjax.jsp"
ADT_ORB_URI = "/ajax/orb.jsp"
ADT_SYSTEM_URI = "/system/system.jsp"
ADT_DEVICE_URI = "/system/device.jsp"
ADT_STATES_URI = "/ajax/currentStates.jsp"
ADT_SYNC_CHECK_URI = "/Ajax/SyncCheckServ"
ADT_TIMEOUT_URI = "/KeepAlive"
# Intervals are all in minutes
ADT_DEFAULT_KEEPALIVE_INTERVAL: int = 5
ADT_DEFAULT_RELOGIN_INTERVAL: int = 120
ADT_MAX_KEEPALIVE_INTERVAL: int = 15
ADT_MIN_RELOGIN_INTERVAL: int = 20
ADT_GATEWAY_STRING = "gateway"

# ADT sets their keepalive to 1 second, so poll a little more often
# than that
ADT_DEFAULT_POLL_INTERVAL = 2.0
ADT_GATEWAY_OFFLINE_POLL_INTERVAL = 90.0
ADT_MAX_RELOGIN_BACKOFF: float = 15.0 * 60.0
ADT_DEFAULT_HTTP_USER_AGENT = {
    "User-Agent": (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/100.0.4896.127 Safari/537.36 Edg/100.0.1185.44"
    )
}

ADT_DEFAULT_HTTP_ACCEPT_HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,"
    "image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7"
}
ADT_DEFAULT_SEC_FETCH_HEADERS = {
    "Sec-Fetch-User": "?1",
    "Sec-Ch-Ua-Mobile": "?0",
    "Sec-Fetch-Site": "same-origin",
    "Sec-Fetch-Mode": "navigate",
    "Upgrade-Insecure-Requests": "1",
}
ADT_OTHER_HTTP_ACCEPT_HEADERS = {
    "Accept": "*/*",
}
ADT_ARM_URI = "/quickcontrol/serv/RunRRACommand"
ADT_ARM_DISARM_URI = "/quickcontrol/armDisarm.jsp"

ADT_SYSTEM_SETTINGS = "/system/settings.jsp"

ADT_DEFAULT_VERSION = "24.0.0-117"

ADT_HTTP_BACKGROUND_URIS = (ADT_ORB_URI, ADT_SYNC_CHECK_URI)
STATE_OK = "OK"
STATE_OPEN = "Open"
STATE_MOTION = "Motion"
STATE_TAMPER = "Tamper"
STATE_ALARM = "Alarm"
STATE_UNKNOWN = "Unknown"
STATE_ONLINE = "Online"

ADT_SENSOR_DOOR = "doorWindow"
ADT_SENSOR_WINDOW = "glass"
ADT_SENSOR_MOTION = "motion"
ADT_SENSOR_SMOKE = "smoke"
ADT_SENSOR_CO = "co"
ADT_SENSOR_ALARM = "alarm"


class ConnectionFailureReason(Enum):
    """Reason for connection failure."""

    NO_FAILURE = 0, "No Failure"
    UNKNOWN = 1, "Unknown Failure"
    ACCOUNT_LOCKED = 2, "Account Locked"
    INVALID_CREDENTIALS = 3, "Invalid Credentials"
    MFA_REQUIRED = 4, "MFA Required"
    SERVICE_UNAVAILABLE = (
        HTTPStatus.SERVICE_UNAVAILABLE.value,
        HTTPStatus.SERVICE_UNAVAILABLE.description,
    )
    TOO_MANY_REQUESTS = (
        HTTPStatus.TOO_MANY_REQUESTS.value,
        HTTPStatus.TOO_MANY_REQUESTS.description,
    )



"""Base Python Class for pyadtpulse."""

import logging
import asyncio
import re
import time
from datetime import datetime
from random import randint
from threading import RLock, Thread
from typing import Union
from warnings import warn

import uvloop
from aiohttp import ClientSession
from bs4 import BeautifulSoup
from yarl import URL

from .alarm_panel import ADT_ALARM_UNKNOWN
from .const import (
    ADT_DEFAULT_HTTP_USER_AGENT,
    ADT_DEFAULT_KEEPALIVE_INTERVAL,
    ADT_DEFAULT_RELOGIN_INTERVAL,
    ADT_GATEWAY_STRING,
    ADT_MAX_KEEPALIVE_INTERVAL,
    ADT_MAX_RELOGIN_BACKOFF,
    ADT_MIN_RELOGIN_INTERVAL,
    ADT_SYNC_CHECK_URI,
    ADT_TIMEOUT_URI,
    DEFAULT_API_HOST,
    ConnectionFailureReason,
)
from .pulse_connection import ADTPulseConnection
from .site import ADTPulseSite
from .util import AuthenticationException, DebugRLock, handle_response

LOG = logging.getLogger(__name__)

SYNC_CHECK_TASK_NAME = "ADT Pulse Sync Check Task"
KEEPALIVE_TASK_NAME = "ADT Pulse Keepalive Task"
RELOGIN_BACKOFF_WARNING_THRESHOLD = 5.0 * 60.0


class PyADTPulse:
    """Base object for ADT Pulse service."""

    __slots__ = (
        "_pulse_connection",
        "_sync_task",
        "_timeout_task",
        "_authenticated",
        "_updates_exist",
        "_session_thread",
        "_attribute_lock",
        "_site",
        "_username",
        "_password",
        "_fingerprint",
        "_login_exception",
        "_relogin_interval",
        "_keepalive_interval",
        "_update_succeded",
        "_detailed_debug_logging",
    )

    @staticmethod
    def _check_keepalive_interval(keepalive_interval: int) -> None:
        if keepalive_interval > ADT_MAX_KEEPALIVE_INTERVAL or keepalive_interval <= 0:
            raise ValueError(
                f"keepalive interval ({keepalive_interval}) must be "
                f"greater than 0 and less than {ADT_MAX_KEEPALIVE_INTERVAL}"
            )

    @staticmethod
    def _check_relogin_interval(relogin_interval: int) -> None:
        if relogin_interval < ADT_MIN_RELOGIN_INTERVAL:
            raise ValueError(
                f"relogin interval ({relogin_interval}) must be "
                f"greater than {ADT_MIN_RELOGIN_INTERVAL}"
            )

    def __init__(
        self,
        username: str,
        password: str,
        fingerprint: str,
        service_host: str = DEFAULT_API_HOST,
        user_agent=ADT_DEFAULT_HTTP_USER_AGENT["User-Agent"],
        websession: ClientSession | None = None,
        do_login: bool = True,
        debug_locks: bool = False,
        keepalive_interval: int = ADT_DEFAULT_KEEPALIVE_INTERVAL,
        relogin_interval: int = ADT_DEFAULT_RELOGIN_INTERVAL,
        detailed_debug_logging: bool = False,
    ):
        """Create a PyADTPulse object.

        Args:
            username (str): Username.
            password (str): Password.
            fingerprint (str): 2FA fingerprint.
            service_host (str, optional): host prefix to use
                         i.e. https://portal.adtpulse.com or
                              https://portal-ca.adtpulse.com
            user_agent (str, optional): User Agent.
                         Defaults to ADT_DEFAULT_HTTP_HEADERS["User-Agent"].
            websession (ClientSession, optional): an initialized
                        aiohttp.ClientSession to use, defaults to None
            do_login (bool, optional): login synchronously when creating object
                            Should be set to False for asynchronous usage
                            and async_login() should be called instead
                            Setting websession will override this
                            and not login
                        Defaults to True
            debug_locks: (bool, optional): use debugging locks
                        Defaults to False
            keepalive_interval (int, optional): number of minutes between
                        keepalive checks, defaults to ADT_DEFAULT_KEEPALIVE_INTERVAL,
                        maxiumum is ADT_MAX_KEEPALIVE_INTERVAL
            relogin_interval (int, optional): number of minutes between relogin checks
                        defaults to ADT_DEFAULT_RELOGIN_INTERVAL,
                        minimum is ADT_MIN_RELOGIN_INTERVAL
            detailed_debug_logging (bool, optional): enable detailed debug logging
        """
        self._init_login_info(username, password, fingerprint)
        self._pulse_connection = ADTPulseConnection(
            service_host,
            session=websession,
            user_agent=user_agent,
            debug_locks=debug_locks,
            detailed_debug_logging=detailed_debug_logging,
        )

        self._sync_task: asyncio.Task | None = None
        self._timeout_task: asyncio.Task | None = None

        # FIXME use thread event/condition, regular condition?
        # defer initialization to make sure we have an event loop
        self._login_exception: BaseException | None = None

        self._updates_exist = asyncio.locks.Event()

        self._session_thread: Thread | None = None
        self._attribute_lock: RLock | DebugRLock
        if not debug_locks:
            self._attribute_lock = RLock()
        else:
            self._attribute_lock = DebugRLock("PyADTPulse._attribute_lock")

        self._site: ADTPulseSite | None = None
        self.keepalive_interval = keepalive_interval
        self.relogin_interval = relogin_interval
        self._detailed_debug_logging = detailed_debug_logging
        self._update_succeded = True

        # authenticate the user
        if do_login and websession is None:
            self.login()

    def _init_login_info(self, username: str, password: str, fingerprint: str) -> None:
        """Initialize login info.

        Raises:
            ValueError: if login parameters are not valid.
        """
        ADTPulseConnection.check_login_parameters(username, password, fingerprint)
        self._username = username
        self._password = password
        self._fingerprint = fingerprint

    def __repr__(self) -> str:
        """Object representation."""
        return f"<{self.__class__.__name__}: {self._username}>"

    # ADTPulse API endpoint is configurable (besides default US ADT Pulse endpoint) to
    # support testing as well as alternative ADT Pulse endpoints such as
    # portal-ca.adtpulse.com

    @property
    def service_host(self) -> str:
        """Get the Pulse host.

        Returns: (str): the ADT Pulse endpoint host
        """
        return self._pulse_connection.service_host

    @service_host.setter
    def service_host(self, host: str) -> None:
        """Override the Pulse host (i.e. to use portal-ca.adpulse.com).

        Args:
            host (str): name of Pulse endpoint host
        """
        ADTPulseConnection.check_service_host(host)
        with self._attribute_lock:
            self._pulse_connection.service_host = host

    def set_service_host(self, host: str) -> None:
        """Backward compatibility for service host property setter."""
        self.service_host = host

    @property
    def username(self) -> str:
        """Get username.

        Returns:
            str: the username
        """
        with self._attribute_lock:
            return self._username

    @property
    def version(self) -> str:
        """Get the ADT Pulse site version.

        Returns:
            str: a string containing the version
        """
        return self._pulse_connection.api_version

    @property
    def relogin_interval(self) -> int:
        """Get re-login interval.

        Returns:
            int: number of minutes to re-login to Pulse
                 0 means disabled
        """
        with self._attribute_lock:
            return self._relogin_interval

    @relogin_interval.setter
    def relogin_interval(self, interval: int | None) -> None:
        """Set re-login interval.

        Args:
            interval (int): The number of minutes between logins.
                            If set to None, resets to ADT_DEFAULT_RELOGIN_INTERVAL

        Raises:
            ValueError: if a relogin interval of less than 10 minutes
                        is specified
        """
        if interval is None:
            interval = ADT_DEFAULT_RELOGIN_INTERVAL
        else:
            self._check_relogin_interval(interval)
        with self._attribute_lock:
            self._relogin_interval = interval
            LOG.debug("relogin interval set to %d", self._relogin_interval)

    @property
    def keepalive_interval(self) -> int:
        """Get the keepalive interval in minutes.

        Returns:
            int: the keepalive interval
        """
        with self._attribute_lock:
            return self._keepalive_interval

    @keepalive_interval.setter
    def keepalive_interval(self, interval: int | None) -> None:
        """Set the keepalive interval in minutes.

        If set to None, resets to ADT_DEFAULT_KEEPALIVE_INTERVAL
        """
        if interval is None:
            interval = ADT_DEFAULT_KEEPALIVE_INTERVAL
        else:
            self._check_keepalive_interval(interval)
        with self._attribute_lock:
            self._keepalive_interval = interval
            LOG.debug("keepalive interval set to %d", self._keepalive_interval)

    @property
    def detailed_debug_logging(self) -> bool:
        """Get the detailed debug logging flag."""
        with self._attribute_lock:
            return self._detailed_debug_logging

    @detailed_debug_logging.setter
    def detailed_debug_logging(self, value: bool) -> None:
        """Set detailed debug logging flag."""
        with self._attribute_lock:
            self._detailed_debug_logging = value
        self._pulse_connection.detailed_debug_logging = value

    @property
    def connection_failure_reason(self) -> ConnectionFailureReason:
        return self._pulse_connection.connection_failure_reason

    async def _update_sites(self, soup: BeautifulSoup) -> None:
        with self._attribute_lock:
            if self._site is None:
                await self._initialize_sites(soup)
                if self._site is None:
                    raise RuntimeError("pyadtpulse could not retrieve site")
            self._site.alarm_control_panel.update_alarm_from_soup(soup)
            self._site.update_zone_from_soup(soup)

    async def _initialize_sites(self, soup: BeautifulSoup) -> None:
        """
        Initializes the sites in the ADT Pulse account.

        Args:
            soup (BeautifulSoup): The parsed HTML soup object.
        """
        # typically, ADT Pulse accounts have only a single site (premise/location)
        single_premise = soup.find("span", {"id": "p_singlePremise"})
        if single_premise:
            site_name = single_premise.text

            # FIXME: this code works, but it doesn't pass the linter
            signout_link = str(
                soup.find("a", {"class": "p_signoutlink"}).get("href")  # type: ignore
            )
            if signout_link:
                m = re.search("networkid=(.+)&", signout_link)
                if m and m.group(1) and m.group(1):
                    site_id = m.group(1)
                    LOG.debug("Discovered site id %s: %s", site_id, site_name)
                    new_site = ADTPulseSite(self._pulse_connection, site_id, site_name)

                    # fetch zones first, so that we can have the status
                    # updated with _update_alarm_status
                    if not await new_site.fetch_devices(None):
                        LOG.error("Could not fetch zones from ADT site")
                    new_site.alarm_control_panel.update_alarm_from_soup(soup)
                    if new_site.alarm_control_panel.status == ADT_ALARM_UNKNOWN:
                        new_site.gateway.is_online = False
                    new_site.update_zone_from_soup(soup)
                    with self._attribute_lock:
                        self._site = new_site
                    return
            else:
                LOG.warning(
                    "Couldn't find site id for %s in %s", site_name, signout_link
                )
        else:
            LOG.error("ADT Pulse accounts with MULTIPLE sites not supported!!!")

    # ...and current network id from:
    # <a id="p_signout1" class="p_signoutlink"
    # href="/myhome/16.0.0-131/access/signout.jsp?networkid=150616za043597&partner=adt"
    # onclick="return flagSignOutInProcess();">
    #
    # ... or perhaps better, just extract all from /system/settings.jsp

    def _get_task_name(self, task, default_name) -> str:
        """
        Get the name of a task.

        Parameters:
            task (Task): The task object.
            default_name (str): The default name to use if the task is None.

        Returns:
            str: The name of the task if it is not None, otherwise the default name
            with a suffix indicating a possible internal error.
        """
        if task is not None:
            return task.get_name()
        return f"{default_name} - possible internal error"

    def _get_sync_task_name(self) -> str:
        return self._get_task_name(self._sync_task, SYNC_CHECK_TASK_NAME)

    def _get_timeout_task_name(self) -> str:
        return self._get_task_name(self._timeout_task, KEEPALIVE_TASK_NAME)

    async def _keepalive_task(self) -> None:
        """
        Asynchronous function that runs a keepalive task to maintain the connection
        with the ADT Pulse cloud.
        """

        async def reset_pulse_cloud_timeout() -> tuple[int, str | None, URL | None]:
            return await self._pulse_connection.async_query(ADT_TIMEOUT_URI, "POST")

        async def update_gateway_device_if_needed() -> None:
            if self.site.gateway.next_update < time.time():
                await self.site.set_device(ADT_GATEWAY_STRING)

        def should_relogin(relogin_interval: int) -> bool:
            return (
                relogin_interval != 0
                and time.time() - self._pulse_connection.last_login_time
                > randint(int(0.75 * relogin_interval), relogin_interval)
            )

        response: str | None
        task_name: str = self._get_task_name(self._timeout_task, KEEPALIVE_TASK_NAME)
        LOG.debug("creating %s", task_name)

        while True:
            relogin_interval = self.relogin_interval
            try:
                await asyncio.sleep(self.keepalive_interval * 60)
                if self._pulse_connection.retry_after > time.time():
                    LOG.debug(
                        "%s: Skipping actions because retry_after > now", task_name
                    )
                    continue
                if not self.is_connected:
                    LOG.debug("%s: Skipping relogin because not connected", task_name)
                    continue
                elif should_relogin(relogin_interval):
                    await self.async_logout()
                    await self._do_login_with_backoff(task_name)
                    continue
                LOG.debug("Resetting timeout")
                code, response, url = await reset_pulse_cloud_timeout()
                if (
                    not handle_response(
                        code,
                        url,
                        logging.WARNING,
                        "Could not reset ADT Pulse cloud timeout",
                    )
                    or response is None
                ):
                    continue
                await update_gateway_device_if_needed()

            except asyncio.CancelledError:
                LOG.debug("%s cancelled", task_name)
                return

    async def _cancel_task(self, task: asyncio.Task | None) -> None:
        """
        Cancel a given asyncio task.

        Args:
            task (asyncio.Task | None): The task to be cancelled.
        """
        if task is None:
            return
        task_name = task.get_name()
        LOG.debug("cancelling %s", task_name)
        try:
            task.cancel()
        except asyncio.CancelledError:
            LOG.debug("%s successfully cancelled", task_name)
            await task

    def _set_update_status(self, value: bool) -> None:
        """Sets update failed, sets updates_exist to notify wait_for_update."""
        with self._attribute_lock:
            self._update_succeded = value
            if self._updates_exist is not None and not self._updates_exist.is_set():
                self._updates_exist.set()

    async def _do_login_with_backoff(self, task_name: str) -> None:
        """
        Performs a logout and re-login process.

        Args:
            None.
        Returns:
            None
        """
        log_level = logging.DEBUG
        login_backoff = 0.0
        login_successful = False

        def compute_login_backoff() -> float:
            if login_backoff == 0.0:
                return self.site.gateway.poll_interval
            return min(ADT_MAX_RELOGIN_BACKOFF, login_backoff * 2.0)

        while not login_successful:
            LOG.log(
                log_level, "%s logging in with backoff %f", task_name, login_backoff
            )
            await asyncio.sleep(login_backoff)
            login_successful = await self.async_login()
            if login_successful:
                if login_backoff != 0.0:
                    self._set_update_status(True)
                return
            # only set flag on first failure
            if login_backoff == 0.0:
                self._set_update_status(False)
            login_backoff = compute_login_backoff()
            if login_backoff > RELOGIN_BACKOFF_WARNING_THRESHOLD:
                log_level = logging.WARNING

    async def _sync_check_task(self) -> None:
        """Asynchronous function that performs a synchronization check task."""

        async def perform_sync_check_query():
            return await self._pulse_connection.async_query(
                ADT_SYNC_CHECK_URI,
                extra_headers={"Sec-Fetch-Mode": "iframe"},
                extra_params={"ts": str(int(time.time() * 1000))},
            )

        task_name = self._get_sync_task_name()
        LOG.debug("creating %s", task_name)

        response_text: str | None = None
        code: int = 200
        last_sync_text = "0-0-0"
        last_sync_check_was_different = False
        url: URL | None = None

        async def validate_sync_check_response() -> bool:
            """
            Validates the sync check response received from the ADT Pulse site.
            Returns:
                bool: True if the sync check response is valid, False otherwise.
            """
            if not handle_response(code, url, logging.ERROR, "Error querying ADT sync"):
                self._set_update_status(False)
                return False
            # this should have already been handled
            if response_text is None:
                LOG.warning("Internal Error: response_text is None")
                return False
            pattern = r"\d+[-]\d+[-]\d+"
            if not re.match(pattern, response_text):
                LOG.warning(
                    "Unexpected sync check format (%s), forcing re-auth",
                    response_text,
                )
                LOG.debug("Received %s from ADT Pulse site", response_text)
                await self.async_logout()
                await self._do_login_with_backoff(task_name)
                return False
            return True

        async def handle_no_updates_exist() -> bool:
            if last_sync_check_was_different:
                if await self.async_update() is False:
                    LOG.debug("Pulse data update from %s failed", task_name)
                    return False
                self._updates_exist.set()
                return True
            else:
                if self.detailed_debug_logging:
                    LOG.debug(
                        "Sync token %s indicates no remote updates to process",
                        response_text,
                    )
            return False

        def handle_updates_exist() -> bool:
            if response_text != last_sync_text:
                LOG.debug("Updates exist: %s, requerying", response_text)
                return True
            return False

        while True:
            try:
                self.site.gateway.adjust_backoff_poll_interval()
                pi = (
                    self.site.gateway.poll_interval
                    if not last_sync_check_was_different
                    else 0.0
                )
                retry_after = self._pulse_connection.retry_after
                if retry_after > time.time():
                    LOG.debug(
                        "%s: Waiting for retry after %s",
                        task_name,
                        datetime.fromtimestamp(retry_after),
                    )
                    self._set_update_status(False)
                    await asyncio.sleep(retry_after - time.time())
                    continue
                await asyncio.sleep(pi)

                code, response_text, url = await perform_sync_check_query()
                if not handle_response(
                    code, url, logging.WARNING, "Error querying ADT sync"
                ):
                    continue
                if response_text is None:
                    LOG.warning("Sync check received no response from ADT Pulse site")
                    continue
                if not await validate_sync_check_response():
                    continue
                if handle_updates_exist():
                    last_sync_check_was_different = True
                    last_sync_text = response_text
                    continue
                if await handle_no_updates_exist():
                    last_sync_check_was_different = False
                    continue
            except asyncio.CancelledError:
                LOG.debug("%s cancelled", task_name)
                return

    def _pulse_session_thread(self) -> None:
        """
        Pulse the session thread.

        Acquires the attribute lock and creates a background thread for the ADT
        Pulse API. The thread runs the synchronous loop `_sync_loop()` until completion.
        Once the loop finishes, the thread is closed, the pulse connection's event loop
        is set to `None`, and the session thread is set to `None`.
        """
        # lock is released in sync_loop()
        self._attribute_lock.acquire()

        LOG.debug("Creating ADT Pulse background thread")
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        loop = asyncio.new_event_loop()
        self._pulse_connection.loop = loop
        loop.run_until_complete(self._sync_loop())

        loop.close()
        self._pulse_connection.loop = None
        self._session_thread = None

    async def _sync_loop(self) -> None:
        """
        Asynchronous function that represents the main loop of the synchronization
        process.

        This function is responsible for executing the synchronization logic. It starts
        by calling the `async_login` method to perform the login operation. After that,
        it releases the `_attribute_lock` to allow other tasks to access the attributes.
        If the login operation was successful, it waits for the `_timeout_task` to
        complete using the `asyncio.wait` function.  If the `_timeout_task` is not set,
        it raises a `RuntimeError` to indicate that background tasks were not created.

        After the waiting process, it enters a while loop that continues as long as the
        `_authenticated` event is set. Inside the loop, it waits for 0.5 seconds using
        the `asyncio.sleep` function. This wait allows the logout process to complete
        before continuing with the synchronization logic.
        """
        result = await self.async_login()
        self._attribute_lock.release()
        if result:
            if self._timeout_task is not None:
                task_list = (self._timeout_task,)
                try:
                    await asyncio.wait(task_list)
                except asyncio.CancelledError:
                    pass
                except Exception as e:  # pylint: disable=broad-except
                    LOG.exception(
                        "Received exception while waiting for ADT Pulse service %s", e
                    )
            else:
                # we should never get here
                raise RuntimeError("Background pyadtpulse tasks not created")
        while self._pulse_connection.authenticated_flag.is_set():
            # busy wait until logout is done
            await asyncio.sleep(0.5)

    def login(self) -> None:
        """Login to ADT Pulse and generate access token.

        Raises:
            AuthenticationException if could not login
        """
        self._attribute_lock.acquire()
        # probably shouldn't be a daemon thread
        self._session_thread = thread = Thread(
            target=self._pulse_session_thread,
            name="PyADTPulse Session",
            daemon=True,
        )
        self._attribute_lock.release()

        self._session_thread.start()
        time.sleep(1)

        # thread will unlock after async_login, so attempt to obtain
        # lock to block current thread until then
        # if it's still alive, no exception
        self._attribute_lock.acquire()
        self._attribute_lock.release()
        if not thread.is_alive():
            raise AuthenticationException(self._username)

    @property
    def attribute_lock(self) -> Union[RLock, DebugRLock]:
        """Get attribute lock for PyADTPulse object.

        Returns:
            RLock: thread Rlock
        """
        return self._attribute_lock

    @property
    def loop(self) -> asyncio.AbstractEventLoop | None:
        """Get event loop.

        Returns:
            Optional[asyncio.AbstractEventLoop]: the event loop object or
                                                 None if no thread is running
        """
        return self._pulse_connection.loop

    async def async_login(self) -> bool:
        """Login asynchronously to ADT.

        Returns: True if login successful
        """
        LOG.debug("Authenticating to ADT Pulse cloud service as %s", self._username)
        await self._pulse_connection.async_fetch_version()

        soup = await self._pulse_connection.async_do_login_query(
            self.username, self._password, self._fingerprint
        )
        if soup is None:
            return False
        # if tasks are started, we've already logged in before
        if self._sync_task is not None or self._timeout_task is not None:
            return True
        await self._update_sites(soup)
        if self._site is None:
            LOG.error("Could not retrieve any sites, login failed")
            await self.async_logout()
            return False

        # since we received fresh data on the status of the alarm, go ahead
        # and update the sites with the alarm status.
        self._timeout_task = asyncio.create_task(
            self._keepalive_task(), name=f"{KEEPALIVE_TASK_NAME}"
        )
        await asyncio.sleep(0)
        return True

    async def async_logout(self) -> None:
        """Logout of ADT Pulse async."""
        LOG.info("Logging %s out of ADT Pulse", self._username)
        if asyncio.current_task() not in (self._sync_task, self._timeout_task):
            await self._cancel_task(self._timeout_task)
            await self._cancel_task(self._sync_task)
            self._timeout_task = self._sync_task = None
        await self._pulse_connection.async_do_logout_query(self.site.id)

    def logout(self) -> None:
        """Log out of ADT Pulse."""
        loop = self._pulse_connection.check_sync(
            "Attempting to call sync logout without sync login"
        )
        sync_thread = self._session_thread

        coro = self.async_logout()
        asyncio.run_coroutine_threadsafe(coro, loop)
        if sync_thread is not None:
            sync_thread.join()

    def _check_update_succeeded(self) -> bool:
        """Check if update succeeded, clears the update event and
        resets _update_succeeded.
        """
        with self._attribute_lock:
            old_update_succeded = self._update_succeded
            self._update_succeded = True
            if self._updates_exist.is_set():
                self._updates_exist.clear()
            return old_update_succeded

    @property
    def updates_exist(self) -> bool:
        """Check if updated data exists.

        Returns:
            bool: True if updated data exists
        """
        with self._attribute_lock:
            if self._sync_task is None:
                loop = self._pulse_connection.loop
                if loop is None:
                    raise RuntimeError(
                        "ADT pulse sync function updates_exist() "
                        "called from async session"
                    )
                coro = self._sync_check_task()
                self._sync_task = loop.create_task(
                    coro, name=f"{SYNC_CHECK_TASK_NAME}: Sync session"
                )
            return self._check_update_succeeded()

    async def wait_for_update(self) -> bool:
        """Wait for update.

        Blocks current async task until Pulse system
        signals an update
        FIXME?: This code probably won't work with multiple waiters.
        """
        with self._attribute_lock:
            if self._sync_task is None:
                coro = self._sync_check_task()
                self._sync_task = asyncio.create_task(
                    coro, name=f"{SYNC_CHECK_TASK_NAME}: Async session"
                )
        if self._updates_exist is None:
            raise RuntimeError("Update event does not exist")

        await self._updates_exist.wait()
        return self._check_update_succeeded()

    @property
    def is_connected(self) -> bool:
        """Check if connected to ADT Pulse.

        Returns:
            bool: True if connected
        """
        return (
            self._pulse_connection.authenticated_flag.is_set()
            and self._pulse_connection.retry_after < time.time()
        )

    async def async_update(self) -> bool:
        """Update ADT Pulse data.

        Returns:
            bool: True if update succeeded.
        """
        LOG.debug("Checking ADT Pulse cloud service for updates")

        # FIXME will have to query other URIs for camera/zwave/etc
        soup = await self._pulse_connection.query_orb(
            logging.INFO, "Error returned from ADT Pulse service check"
        )
        if soup is not None:
            await self._update_sites(soup)
            return True

        return False

    def update(self) -> bool:
        """Update ADT Pulse data.

        Returns:
            bool: True on success
        """
        coro = self.async_update()
        return asyncio.run_coroutine_threadsafe(
            coro,
            self._pulse_connection.check_sync(
                "Attempting to run sync update from async login"
            ),
        ).result()

    @property
    def sites(self) -> list[ADTPulseSite]:
        """Return all sites for this ADT Pulse account."""
        warn(
            "multiple sites being removed, use pyADTPulse.site instead",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        with self._attribute_lock:
            if self._site is None:
                raise RuntimeError(
                    "No sites have been retrieved, have you logged in yet?"
                )
            return [self._site]

    @property
    def site(self) -> ADTPulseSite:
        """Return the site associated with the Pulse login."""
        with self._attribute_lock:
            if self._site is None:
                raise RuntimeError(
                    "No sites have been retrieved, have you logged in yet?"
                )
            return self._site



"""ADT Pulse zone info."""
import logging
from collections import UserDict
from dataclasses import dataclass
from datetime import datetime
from typing import TypedDict

ADT_NAME_TO_DEFAULT_TAGS = {
    "Door": ("sensor", "doorWindow"),
    "Window": ("sensor", "doorWindow"),
    "Motion": ("sensor", "motion"),
    "Glass": ("sensor", "glass"),
    "Gas": ("sensor", "co"),
    "Carbon": ("sensor", "co"),
    "Smoke": ("sensor", "smoke"),
    "Flood": ("sensor", "flood"),
    "Floor": ("sensor", "flood"),
    "Moisture": ("sensor", "flood"),
}

LOG = logging.getLogger(__name__)


@dataclass(slots=True)
class ADTPulseZoneData:
    """Data for an ADT Pulse zone.

    Fields:
        name (str): the zone name
        id_ (str): zone name in ADT Pulse (Note, not id as this is a reserved word)
        tags (Tuple): sensor and type(s)
        status (str): sensor status (i.e. Online, Low Battery, etc.)
        state (str): sensor state (i.e. Opened, Closed, etc)
        timestamp (datetime): timestamp of last activity

    Will set unknown type defaults to all but name and id_
    """

    name: str
    id_: str
    tags: tuple = ADT_NAME_TO_DEFAULT_TAGS["Window"]
    status: str = "Unknown"
    state: str = "Unknown"
    last_activity_timestamp: int = 0


class ADTPulseFlattendZone(TypedDict):
    """Represent ADTPulseZones as a "flattened" dictionary.

    Fields:
        zone (int): the zone id
        name (str): the zone name
        id_ (str): zone name in ADT Pulse (Note, not id as this is a reserved word)
        tags (Tuple): sensor and type(s)
        status (str): sensor status (i.e. Online, Low Battery, etc.)
        state (str): sensor state (i.e. Opened, Closed, etc)
        timestamp (datetime): timestamp of last activity
    """

    zone: int
    name: str
    id_: str
    tags: tuple
    status: str
    state: str
    last_activity_timestamp: int


class ADTPulseZones(UserDict):
    """Dictionary containing ADTPulseZoneData with zone as the key."""

    @staticmethod
    def _check_value(value: ADTPulseZoneData) -> None:
        if not isinstance(value, ADTPulseZoneData):
            raise ValueError("ADT Pulse zone data must be of type ADTPulseZoneData")

    @staticmethod
    def _check_key(key: int) -> None:
        if not isinstance(key, int):
            raise ValueError("ADT Pulse Zone must be an integer")

    def __getitem__(self, key: int) -> ADTPulseZoneData:
        """Get a Zone.

        Args:
            key (int): zone id

        Returns:
            ADTPulseZoneData: zone data
        """
        return super().__getitem__(key)

    def _get_zonedata(self, key: int) -> ADTPulseZoneData:
        self._check_key(key)
        result: ADTPulseZoneData = self.data[key]
        self._check_value(result)
        return result

    def __setitem__(self, key: int, value: ADTPulseZoneData) -> None:
        """Validate types and sets defaults for ADTPulseZones.

        ADTPulseZoneData.id_ and name will be set to generic defaults if not set

        Raises:
            ValueError: if key is not an int or value is not ADTPulseZoneData
        """
        self._check_key(key)
        self._check_value(value)
        if not value.id_:
            value.id_ = "sensor-" + str(key)
        if not value.name:
            value.name = "Sensor for Zone " + str(key)
        super().__setitem__(key, value)

    def update_status(self, key: int, status: str) -> None:
        """Update zone status.

        Args:
            key (int): zone id to change
            status (str): status to set
        """ """"""
        temp = self._get_zonedata(key)
        temp.status = status
        self.__setitem__(key, temp)

    def update_state(self, key: int, state: str) -> None:
        """Update zone state.

        Args:
            key (int): zone id to change
            state (str): state to set
        """
        temp = self._get_zonedata(key)
        temp.state = state
        self.__setitem__(key, temp)

    def update_last_activity_timestamp(self, key: int, dt: datetime) -> None:
        """Update timestamp.

        Args:
            key (int): zone id to change
            dt (datetime): timestamp to set
        """
        temp = self._get_zonedata(key)
        temp.last_activity_timestamp = int(dt.timestamp())
        self.__setitem__(key, temp)

    def update_device_info(
        self,
        key: int,
        state: str,
        status: str = "Online",
        last_activity: datetime = datetime.now(),
    ) -> None:
        """Update the device info.

        Convenience method to update all common device info
        at once.

        Args:
            key (int): zone id
            state (str): state
            status (str, optional): status. Defaults to "Online".
            last_activity (datetime, optional): last_activity_datetime.
                Defaults to datetime.now().
        """
        temp = self._get_zonedata(key)
        temp.state = state
        temp.status = status
        temp.last_activity_timestamp = int(last_activity.timestamp())
        self.__setitem__(key, temp)

    def flatten(self) -> list[ADTPulseFlattendZone]:
        """Flattens ADTPulseZones into a list of ADTPulseFlattenedZones.

        Returns:
            List[ADTPulseFlattendZone]
        """
        result: list[ADTPulseFlattendZone] = []
        for k, i in self.items():
            if not isinstance(i, ADTPulseZoneData):
                raise ValueError("Invalid Zone data in ADTPulseZones")
            result.append(
                {
                    "zone": k,
                    "name": i.name,
                    "id_": i.id_,
                    "tags": i.tags,
                    "status": i.status,
                    "state": i.state,
                    "last_activity_timestamp": i.last_activity_timestamp,
                }
            )
        return result

    def update_zone_attributes(self, dev_attr: dict[str, str]) -> None:
        """Update zone attributes."""
        d_name = dev_attr.get("name", "Unknown")
        d_type = dev_attr.get("type_model", "Unknown")
        d_zone = dev_attr.get("zone", "Unknown")
        d_status = dev_attr.get("status", "Unknown")

        if d_zone != "Unknown":
            tags = None
            for search_term, default_tags in ADT_NAME_TO_DEFAULT_TAGS.items():
                # convert to uppercase first
                if search_term.upper() in d_type.upper():
                    tags = default_tags
                    break
            if not tags:
                LOG.warning(
                    "Unknown sensor type for '%s', defaulting to doorWindow", d_type
                )
                tags = ("sensor", "doorWindow")
            LOG.debug(
                "Retrieved sensor %s id: sensor-%s Status: %s, tags %s",
                d_name,
                d_zone,
                d_status,
                tags,
            )
            if "Unknown" in (d_name, d_status, d_zone) or not d_zone.isdecimal():
                LOG.debug("Zone data incomplete, skipping...")
            else:
                tmpzone = ADTPulseZoneData(d_name, f"sensor-{d_zone}", tags, d_status)
                self.update({int(d_zone): tmpzone})
        else:
            LOG.debug(
                "Skipping incomplete zone name: %s, zone: %s status: %s",
                d_name,
                d_zone,
                d_status,
            )



# Initially Generated by CodiumAI

# Dependencies:
# pip install pytest-mock
import pytest

from pyadtpulse.alarm_panel import ADT_ALARM_AWAY, ADTPulseAlarmPanel


class TestADTPulseAlarmPanel:
    # ADTPulseAlarmPanel object is created with default values

    # ADTPulseAlarmPanel attributes are set correctly
    def test_set_attributes(self):
        alarm_panel = ADTPulseAlarmPanel()
        alarm_attributes = {
            "type_model": "Model",
            "manufacturer_provider": "Manufacturer",
            "status": "Online",
        }
        alarm_panel.set_alarm_attributes(alarm_attributes)
        assert alarm_panel.model == "Model"
        assert alarm_panel.manufacturer == "Manufacturer"
        assert alarm_panel.online == True

    # ADTPulseAlarmPanel is able to handle connection errors
    @pytest.mark.asyncio
    async def test_handle_connection_errors(self, mocker):
        # Mock ADTPulseConnection
        mock_connection = mocker.Mock()
        mock_connection.async_query.side_effect = ConnectionError

        # Create ADTPulseAlarmPanel instance
        alarm_panel = ADTPulseAlarmPanel()

        # Call the method that should handle connection errors
        result = await alarm_panel._arm(mock_connection, ADT_ALARM_AWAY, False)

        # Verify that the method returns False
        assert result == False

    # ADTPulseAlarmPanel is able to handle timeouts
    @pytest.mark.asyncio
    async def test_handle_timeouts(self, mocker):
        # Mock ADTPulseConnection
        mock_connection = mocker.Mock()
        mock_connection.async_query.side_effect = TimeoutError

        # Create ADTPulseAlarmPanel instance
        alarm_panel = ADTPulseAlarmPanel()

        # Call the method that should handle timeouts
        result = await alarm_panel._arm(mock_connection, ADT_ALARM_AWAY, False)

        # Verify that the method returns False
        assert result == False

    # ADTPulseAlarmPanel is able to handle unexpected HTML soup
    @pytest.mark.asyncio
    async def test_handle_unexpected_html_soup(self, mocker):
        # Mock ADTPulseConnection
        mock_connection = mocker.Mock()
        mock_connection.async_query.return_value = '<html><body><div class="p_armDisarmWrapper"><div>Error</div></div></body></html>'

        # Create ADTPulseAlarmPanel instance
        alarm_panel = ADTPulseAlarmPanel()

        # Call the method that should handle unexpected HTML soup
        result = await alarm_panel._arm(mock_connection, ADT_ALARM_AWAY, False)

        # Verify that the method returns False
        assert result == False



"""Module representing an ADT Pulse Site."""
import logging
import re
from asyncio import Task, create_task, gather, get_event_loop, run_coroutine_threadsafe
from datetime import datetime
from threading import RLock
from time import time
from typing import Union
from warnings import warn

# import dateparser
from bs4 import BeautifulSoup, ResultSet

from .alarm_panel import ADTPulseAlarmPanel
from .const import ADT_DEVICE_URI, ADT_GATEWAY_STRING, ADT_SYSTEM_URI
from .gateway import ADTPulseGateway
from .pulse_connection import ADTPulseConnection
from .util import DebugRLock, make_soup, parse_pulse_datetime, remove_prefix
from .zones import ADTPulseFlattendZone, ADTPulseZones

LOG = logging.getLogger(__name__)

SECURITY_PANEL_ID = "1"
SECURITY_PANEL_NAME = "Security Panel"


class ADTPulseSite:
    """Represents an individual ADT Pulse site."""

    __slots__ = (
        "_pulse_connection",
        "_id",
        "_name",
        "_last_updated",
        "_alarm_panel",
        "_zones",
        "_site_lock",
        "_gateway",
    )

    def __init__(self, pulse_connection: ADTPulseConnection, site_id: str, name: str):
        """Initialize.

        Args:
            adt_service (PyADTPulse): a PyADTPulse object
            site_id (str): site ID
            name (str): site name
        """
        self._pulse_connection = pulse_connection
        self._id = site_id
        self._name = name
        self._last_updated: int = 0
        self._zones = ADTPulseZones()
        self._site_lock: RLock | DebugRLock
        if isinstance(self._pulse_connection._attribute_lock, DebugRLock):
            self._site_lock = DebugRLock("ADTPulseSite._site_lock")
        else:
            self._site_lock = RLock()
        self._alarm_panel = ADTPulseAlarmPanel()
        self._gateway = ADTPulseGateway()

    @property
    def id(self) -> str:
        """Get site id.

        Returns:
            str: the site id
        """
        return self._id

    @property
    def name(self) -> str:
        """Get site name.

        Returns:
            str: the site name
        """
        return self._name

    # FIXME: should this actually return if the alarm is going off!?  How do we
    # return state that shows the site is compromised??

    @property
    def last_updated(self) -> int:
        """Return time site last updated.

        Returns:
            int: the time site last updated as datetime
        """
        with self._site_lock:
            return self._last_updated

    @property
    def site_lock(self) -> Union[RLock, DebugRLock]:
        """Get thread lock for site data.

        Not needed for async

        Returns:
            RLock: thread RLock
        """
        return self._site_lock

    def arm_home(self, force_arm: bool = False) -> bool:
        """Arm system home."""
        if self.alarm_control_panel is None:
            raise RuntimeError("Cannot arm system home, no control panels exist")
        return self.alarm_control_panel.arm_home(
            self._pulse_connection, force_arm=force_arm
        )

    def arm_away(self, force_arm: bool = False) -> bool:
        """Arm system away."""
        if self.alarm_control_panel is None:
            raise RuntimeError("Cannot arm system away, no control panels exist")
        return self.alarm_control_panel.arm_away(
            self._pulse_connection, force_arm=force_arm
        )

    def disarm(self) -> bool:
        """Disarm system."""
        if self.alarm_control_panel is None:
            raise RuntimeError("Cannot disarm system, no control panels exist")
        return self.alarm_control_panel.disarm(self._pulse_connection)

    async def async_arm_home(self, force_arm: bool = False) -> bool:
        """Arm system home async."""
        if self.alarm_control_panel is None:
            raise RuntimeError("Cannot arm system home, no control panels exist")
        return await self.alarm_control_panel.async_arm_home(
            self._pulse_connection, force_arm=force_arm
        )

    async def async_arm_away(self, force_arm: bool = False) -> bool:
        """Arm system away async."""
        if self.alarm_control_panel is None:
            raise RuntimeError("Cannot arm system away, no control panels exist")
        return await self.alarm_control_panel.async_arm_away(
            self._pulse_connection, force_arm=force_arm
        )

    async def async_disarm(self) -> bool:
        """Disarm system async."""
        if self.alarm_control_panel is None:
            raise RuntimeError("Cannot disarm system, no control panels exist")
        return await self.alarm_control_panel.async_disarm(self._pulse_connection)

    @property
    def zones(self) -> list[ADTPulseFlattendZone] | None:
        """Return all zones registered with the ADT Pulse account.

        (cached copy of last fetch)
        See Also fetch_zones()
        """
        with self._site_lock:
            if not self._zones:
                raise RuntimeError("No zones exist")
            return self._zones.flatten()

    @property
    def zones_as_dict(self) -> ADTPulseZones | None:
        """Return zone information in dictionary form.

        Returns:
            ADTPulseZones: all zone information
        """
        with self._site_lock:
            if not self._zones:
                raise RuntimeError("No zones exist")
            return self._zones

    @property
    def alarm_control_panel(self) -> ADTPulseAlarmPanel:
        """Return the alarm panel object for the site.

        Returns:
            Optional[ADTPulseAlarmPanel]: the alarm panel object
        """
        return self._alarm_panel

    @property
    def gateway(self) -> ADTPulseGateway:
        """Get gateway device object.

        Returns:
            ADTPulseGateway: Gateway device
        """
        return self._gateway

    @property
    def history(self):
        """Return log of history for this zone (NOT IMPLEMENTED)."""
        raise NotImplementedError

        #        status_orb = summary_html_soup.find('canvas', {'id': 'ic_orb'})
        #        if status_orb:
        #            self._status = status_orb['orb']
        #            LOG.warning(status_orb)
        #            LOG.debug("Alarm status = %s", self._status)
        #        else:
        #            LOG.error("Failed to find alarm status in ADT summary!")

        # if we should also update the zone details, force a fresh fetch
        # of data from ADT Pulse

    async def _get_device_attributes(self, device_id: str) -> dict[str, str] | None:
        """
        Retrieves the attributes of a device.

        Args:
            device_id (str): The ID of the device to retrieve attributes for.

        Returns:
            Optional[dict[str, str]]: A dictionary of attribute names and their
                corresponding values,
                or None if the device response soup is None.
        """
        result: dict[str, str] = {}
        if device_id == ADT_GATEWAY_STRING:
            device_response = await self._pulse_connection.async_query(
                "/system/gateway.jsp", timeout=10
            )
        else:
            device_response = await self._pulse_connection.async_query(
                ADT_DEVICE_URI, extra_params={"id": device_id}
            )
        device_response_soup = make_soup(
            device_response[0],
            device_response[1],
            device_response[2],
            logging.DEBUG,
            "Failed loading device attributes from ADT Pulse service",
        )
        if device_response_soup is None:
            return None
        for dev_info_row in device_response_soup.find_all(
            "td", {"class", "InputFieldDescriptionL"}
        ):
            identity_text = (
                str(dev_info_row.get_text())
                .lower()
                .strip()
                .rstrip(":")
                .replace(" ", "_")
                .replace("/", "_")
            )
            sibling = dev_info_row.find_next_sibling()
            if not sibling:
                value = "Unknown"
            else:
                value = str(sibling.get_text()).strip()
            result.update({identity_text: value})
        return result

    async def set_device(self, device_id: str) -> None:
        """
        Sets the device attributes for the given device ID.

        Args:
            device_id (str): The ID of the device.
        """
        dev_attr = await self._get_device_attributes(device_id)
        if dev_attr is None:
            return
        if device_id == ADT_GATEWAY_STRING:
            self._gateway.set_gateway_attributes(dev_attr)
            return
        if device_id == SECURITY_PANEL_ID:
            self._alarm_panel.set_alarm_attributes(dev_attr)
            return
        if device_id.isdigit():
            self._zones.update_zone_attributes(dev_attr)
        else:
            LOG.debug("Zone %s is not an integer, skipping", device_id)

    async def fetch_devices(self, soup: BeautifulSoup | None) -> bool:
        """
        Fetches the devices from the given BeautifulSoup object and updates
        the zone attributes.

        Args:
            soup (Optional[BeautifulSoup]): The BeautifulSoup object containing
                the devices.

        Returns:
            bool: True if the devices were fetched and zone attributes were updated
                successfully, False otherwise.
        """
        regex_device = r"goToUrl\('device.jsp\?id=(\d*)'\);"
        task_list: list[Task] = []
        zone_id: str | None = None

        def add_zone_from_row(row_tds: ResultSet) -> str | None:
            """Adds a zone from a bs4 row.

            Returns None if successful, otherwise the zone ID if present.
            """
            zone_id: str | None = None
            if row_tds and len(row_tds) > 4:
                zone_name: str = row_tds[1].get_text().strip()
                zone_id = row_tds[2].get_text().strip()
                zone_type: str = row_tds[4].get_text().strip()
                zone_status: str = row_tds[0].find("canvas").get("title").strip()
                if (
                    zone_id is not None
                    and zone_id.isdecimal()
                    and zone_name
                    and zone_type
                ):
                    self._zones.update_zone_attributes(
                        {
                            "name": zone_name,
                            "zone": zone_id,
                            "type_model": zone_type,
                            "status": zone_status,
                        }
                    )
                    return None
            return zone_id

        def check_panel_or_gateway(
            regex_device: str,
            device_name: str,
            zone_id: str | None,
            on_click_value_text: str,
        ) -> Task | None:
            result = re.findall(regex_device, on_click_value_text)
            if result:
                device_id = result[0]
                if device_id == SECURITY_PANEL_ID or device_name == SECURITY_PANEL_NAME:
                    return create_task(self.set_device(device_id))
                if zone_id and zone_id.isdecimal():
                    return create_task(self.set_device(device_id))
            LOG.debug("Skipping %s as it doesn't have an ID", device_name)
            return None

        if not soup:
            response = await self._pulse_connection.async_query(ADT_SYSTEM_URI)
            soup = make_soup(
                response[0],
                response[1],
                response[2],
                logging.WARNING,
                "Failed loading zone status from ADT Pulse service",
            )
            if not soup:
                return False
        with self._site_lock:
            for row in soup.find_all("tr", {"class": "p_listRow", "onclick": True}):
                device_name = row.find("a").get_text()
                row_tds = row.find_all("td")
                zone_id = add_zone_from_row(row_tds)
                if zone_id is None:
                    continue
                on_click_value_text = row.get("onclick")
                if (
                    on_click_value_text in ("goToUrl('gateway.jsp');", "Gateway")
                    or device_name == "Gateway"
                ):
                    task_list.append(create_task(self.set_device(ADT_GATEWAY_STRING)))
                elif (
                    result := check_panel_or_gateway(
                        regex_device,
                        device_name,
                        zone_id,
                        on_click_value_text,
                    )
                ) is not None:
                    task_list.append(result)

        await gather(*task_list)
        self._last_updated = int(time())
        return True

    async def _async_update_zones_as_dict(
        self, soup: BeautifulSoup | None
    ) -> ADTPulseZones | None:
        """Update zone status information asynchronously.

        Returns:
            ADTPulseZones: a dictionary of zones with status
            None if an error occurred
        """
        with self._site_lock:
            if self._zones is None:
                self._site_lock.release()
                raise RuntimeError("No zones exist")
            LOG.debug("fetching zones for site %s", self._id)
            if not soup:
                # call ADT orb uri
                soup = await self._pulse_connection.query_orb(
                    logging.WARNING, "Could not fetch zone status updates"
                )
            if soup is None:
                return None
            return self.update_zone_from_soup(soup)

    def update_zone_from_soup(self, soup: BeautifulSoup) -> ADTPulseZones | None:
        """
        Updates the zone information based on the provided BeautifulSoup object.

        Args:
            soup (BeautifulSoup): The BeautifulSoup object containing the parsed HTML.

        Returns:
            Optional[ADTPulseZones]: The updated ADTPulseZones object, or None if
            no zones exist.
        """
        # parse ADT's convulated html to get sensor status
        with self._site_lock:
            gateway_online = False
            for row in soup.find_all("tr", {"class": "p_listRow"}):
                temp = row.find("span", {"class": "devStatIcon"})
                if temp is None:
                    break
                last_update = datetime(1970, 1, 1)
                try:
                    last_update = parse_pulse_datetime(
                        remove_prefix(temp.get("title"), "Last Event:")
                    )
                except ValueError:
                    last_update = datetime(1970, 1, 1)
                # name = row.find("a", {"class": "p_deviceNameText"}).get_text()
                temp = row.find("span", {"class": "p_grayNormalText"})
                if temp is None:
                    break
                zone = int(
                    remove_prefix(
                        temp.get_text(),
                        "Zone\xa0",
                    )
                )
                state = remove_prefix(
                    row.find("canvas", {"class": "p_ic_icon_device"}).get("icon"),
                    "devStat",
                )
                temp_status = row.find("td", {"class": "p_listRow"}).find_next(
                    "td", {"class": "p_listRow"}
                )

                status = "Unknown"
                if temp_status is not None:
                    temp_status = temp_status.get_text()
                    if temp_status is not None:
                        temp_status = str(temp_status.replace("\xa0", ""))
                        if temp_status.startswith("Trouble"):
                            trouble_code = str(temp_status).split()
                            if len(trouble_code) > 1:
                                status = " ".join(trouble_code[1:])
                            else:
                                status = "Unknown trouble code"
                        else:
                            status = "Online"

                # parse out last activity (required dealing with "Yesterday 1:52 PM")
                #           last_activity = time.time()

                # id:    [integer]
                # name:  device name
                # tags:  sensor,[doorWindow,motion,glass,co,fire]
                # timestamp: timestamp of last activity
                # state: OK (device okay)
                #        Open (door/window opened)
                #        Motion (detected motion)
                #        Tamper (glass broken or device tamper)
                #        Alarm (detected CO/Smoke)
                #        Unknown (device offline)

                # update device state from ORB info
                if not self._zones:
                    LOG.warning("No zones exist")
                    return None
                if state != "Unknown":
                    gateway_online = True
                self._zones.update_device_info(zone, state, status, last_update)
                LOG.debug(
                    "Set zone %d - to %s, status %s with timestamp %s",
                    zone,
                    state,
                    status,
                    last_update,
                )
            self._gateway.is_online = gateway_online
            self._last_updated = int(time())
            return self._zones

    async def _async_update_zones(self) -> list[ADTPulseFlattendZone] | None:
        """Update zones asynchronously.

        Returns:
            List[ADTPulseFlattendZone]: a list of zones with their status

            None on error
        """
        with self._site_lock:
            if not self._zones:
                return None
            zonelist = await self._async_update_zones_as_dict(None)
            if not zonelist:
                return None
            return zonelist.flatten()

    def update_zones(self) -> list[ADTPulseFlattendZone] | None:
        """Update zone status information.

        Returns:
            Optional[List[ADTPulseFlattendZone]]: a list of zones with status
        """
        coro = self._async_update_zones()
        return run_coroutine_threadsafe(coro, get_event_loop()).result()

    @property
    def updates_may_exist(self) -> bool:
        """Query whether updated sensor data exists.

        Deprecated, use method on pyADTPulse object instead
        """
        # FIXME: this should actually capture the latest version
        # and compare if different!!!
        # ...this doesn't actually work if other components are also checking
        #  if updates exist
        warn(
            "updates_may_exist on site object is deprecated, "
            "use method on pyADTPulse object instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return False

    async def async_update(self) -> bool:
        """Force update site/zone data async with current data.

        Deprecated, use method on pyADTPulse object instead
        """
        warn(
            "updating zones from site object is deprecated, "
            "use method on pyADTPulse object instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return False

    def update(self) -> bool:
        """Force update site/zones with current data.

        Deprecated, use method on pyADTPulse object instead
        """
        warn(
            "updating zones from site object is deprecated, "
            "use method on pyADTPulse object instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return False



# Generated by CodiumAI
from datetime import datetime, time, timedelta
from ipaddress import IPv4Address

import pytest

from pyadtpulse.const import (
    ADT_DEFAULT_POLL_INTERVAL,
    ADT_GATEWAY_OFFLINE_POLL_INTERVAL,
)
from pyadtpulse.gateway import ADTPulseGateway


class TestADTPulseGateway:
    # create an instance of ADTPulseGateway with default values
    def test_create_instance_with_default_values(self):
        gateway = ADTPulseGateway()
        assert gateway.manufacturer == "Unknown"
        assert gateway._status_text == "OFFLINE"
        assert gateway._current_poll_interval == ADT_DEFAULT_POLL_INTERVAL
        assert gateway._initial_poll_interval == ADT_DEFAULT_POLL_INTERVAL
        assert gateway.model is None
        assert gateway.serial_number is None
        assert gateway.next_update == 0
        assert gateway.last_update == 0
        assert gateway.firmware_version is None
        assert gateway.hardware_version is None
        assert gateway.primary_connection_type is None
        assert gateway.broadband_connection_status is None
        assert gateway.cellular_connection_status is None
        assert gateway.cellular_connection_signal_strength == 0.0
        assert gateway.broadband_lan_ip_address is None
        assert gateway.broadband_lan_mac is None
        assert gateway.device_lan_ip_address is None
        assert gateway.device_lan_mac is None
        assert gateway.router_lan_ip_address is None
        assert gateway.router_wan_ip_address is None

    # set is_online to True and check that it returns True
    def test_set_is_online_true(self):
        gateway = ADTPulseGateway()
        gateway.is_online = True
        assert gateway.is_online is True

    # set is_online to False and check that it returns False
    def test_set_is_online_false(self):
        gateway = ADTPulseGateway()
        gateway.is_online = False
        assert gateway.is_online is False

    # set poll_interval to a valid value and check that it returns the same value
    def test_set_poll_interval_valid_value(self):
        gateway = ADTPulseGateway()
        gateway.poll_interval = 10.0
        assert gateway.poll_interval == 10.0

    # set poll_interval to None and check that it returns ADT_DEFAULT_POLL_INTERVAL
    def test_set_poll_interval_none(self):
        gateway = ADTPulseGateway()
        gateway.poll_interval = None
        assert gateway.poll_interval == ADT_DEFAULT_POLL_INTERVAL

    # set gateway attributes using set_gateway_attributes method and check that
    # they are set correctly
    def test_set_gateway_attributes(self):
        gateway = ADTPulseGateway()
        attributes = {
            "manufacturer": "ADT",
            "_status_text": "ONLINE",
            "model": "1234",
            "serial_number": "5678",
            "next_update": "Today\xa09:03\xa0AM",
            "last_update": "Yesterday\xa011:55\xa0PM",
            "firmware_version": "1.0",
            "hardware_version": "2.0",
            "primary_connection_type": "Ethernet",
            "broadband_connection_status": "Connected",
            "cellular_connection_status": "Disconnected",
            "cellular_connection_signal_strength": 3.5,
            "broadband_lan_ip_address": IPv4Address("192.168.1.1"),
            "broadband_lan_mac": "00:11:22:33:44:55",
            "device_lan_ip_address": IPv4Address("192.168.1.2"),
            "device_lan_mac": "AA:BB:CC:DD:EE:FF",
            "router_lan_ip_address": IPv4Address("192.168.1.3"),
            "router_wan_ip_address": IPv4Address("10.0.0.1"),
        }
        gateway.set_gateway_attributes(attributes)
        assert gateway.manufacturer == "ADT"
        assert gateway.model == "1234"
        assert gateway.serial_number == "5678"
        now = datetime.now()
        yesterday = now - timedelta(days=1)
        assert gateway.next_update == int(
            datetime.combine(now.date(), time(9, 3)).timestamp()
        )
        assert gateway.last_update == int(
            datetime.combine(yesterday.date(), time(23, 55)).timestamp()
        )
        assert gateway.firmware_version == "1.0"
        assert gateway.hardware_version == "2.0"
        assert gateway.primary_connection_type == "Ethernet"
        assert gateway.broadband_connection_status == "Connected"
        assert gateway.cellular_connection_status == "Disconnected"
        assert gateway.cellular_connection_signal_strength == 3.5
        assert gateway.broadband_lan_ip_address == IPv4Address("192.168.1.1")
        assert gateway.broadband_lan_mac == "00:11:22:33:44:55"
        assert gateway.device_lan_ip_address == IPv4Address("192.168.1.2")
        assert gateway.device_lan_mac == "AA:BB:CC:DD:EE:FF"
        assert gateway.router_lan_ip_address == IPv4Address("192.168.1.3")
        assert gateway.router_wan_ip_address == IPv4Address("10.0.0.1")

    # set poll_interval to a negative value and check that it raises a ValueError
    def test_set_poll_interval_negative_value(self):
        gateway = ADTPulseGateway()
        with pytest.raises(ValueError):
            gateway.poll_interval = -10.0

    # set gateway attributes with invalid IP address and check that it
    # sets the attribute to None
    def test_set_gateway_attributes_invalid_ip_address(self):
        gateway = ADTPulseGateway()
        attributes = {
            "broadband_lan_ip_address": "invalid_ip_address",
            "device_lan_ip_address": "invalid_ip_address",
            "router_lan_ip_address": "invalid_ip_address",
            "router_wan_ip_address": "invalid_ip_address",
        }
        gateway.set_gateway_attributes(attributes)
        assert gateway.broadband_lan_ip_address is None
        assert gateway.device_lan_ip_address is None
        assert gateway.router_lan_ip_address is None
        assert gateway.router_wan_ip_address is None

    # set gateway attributes with invalid datetime and check that it
    # sets the attribute to None
    def test_set_gateway_attributes_invalid_datetime(self):
        gateway = ADTPulseGateway()
        attributes = {
            "next_update": "invalid_datetime",
            "last_update": "invalid_datetime",
        }
        gateway.set_gateway_attributes(attributes)
        assert gateway.next_update is None
        assert gateway.last_update is None

    # set is_online to True when it is already True and check that it does not
    # change the status
    def test_set_is_online_true_already_true(self):
        gateway = ADTPulseGateway()
        gateway.is_online = True
        gateway.is_online = True
        assert gateway.is_online is True

    # set is_online to False when it is already False and check that it does
    # not change the status
    def test_set_is_online_false_already_false(self):
        gateway = ADTPulseGateway()
        gateway.is_online = False
        gateway.is_online = False
        assert gateway.is_online is False

    # Check that adjust_backoff_poll_interval method sets the current poll
    # interval correctly
    def test_adjust_backoff_poll_interval_fixed(self):
        gateway = ADTPulseGateway()
        gateway.poll_interval = ADT_DEFAULT_POLL_INTERVAL
        gateway.adjust_backoff_poll_interval()
        assert gateway._current_poll_interval == ADT_DEFAULT_POLL_INTERVAL * 2
        gateway.is_online = True
        gateway.adjust_backoff_poll_interval()
        assert gateway._current_poll_interval == ADT_DEFAULT_POLL_INTERVAL
        gateway.is_online = False
        gateway.adjust_backoff_poll_interval()
        assert gateway._current_poll_interval == ADT_DEFAULT_POLL_INTERVAL
        gateway.adjust_backoff_poll_interval()
        assert gateway._current_poll_interval == ADT_DEFAULT_POLL_INTERVAL * 2
        gateway.adjust_backoff_poll_interval()
        assert gateway._current_poll_interval == ADT_DEFAULT_POLL_INTERVAL * 4
        gateway.is_online = True
        gateway.adjust_backoff_poll_interval()
        assert gateway._current_poll_interval == ADT_DEFAULT_POLL_INTERVAL

    # check that is_online setter changes the poll interval correctly
    def test_is_online_setter_changes_poll_interval(self):
        gateway = ADTPulseGateway()
        gateway.is_online = True
        assert gateway._status_text == "ONLINE"
        assert gateway._current_poll_interval == ADT_DEFAULT_POLL_INTERVAL
        gateway.is_online = False
        assert gateway._status_text == "OFFLINE"
        assert gateway._current_poll_interval == ADT_GATEWAY_OFFLINE_POLL_INTERVAL

    # check that poll_interval setter changes the poll interval correctly
    def test_poll_interval_setter_changes_poll_interval(self):
        gateway = ADTPulseGateway()
        gateway.poll_interval = 60.0
        assert gateway._initial_poll_interval == 60.0
        assert gateway._current_poll_interval == 60.0
        gateway.poll_interval = None
        assert gateway._initial_poll_interval == ADT_DEFAULT_POLL_INTERVAL
        assert gateway._current_poll_interval == ADT_DEFAULT_POLL_INTERVAL

    # check that set_gateway_attributes method sets None for empty strings
    def test_set_gateway_attributes_sets_none_for_empty_strings(self):
        gateway = ADTPulseGateway()
        attributes = {
            "manufacturer": "",
            "model": "",
            "serial_number": "",
            "firmware_version": "",
            "hardware_version": "",
            "primary_connection_type": "",
            "broadband_connection_status": "",
            "cellular_connection_status": "",
            "broadband_lan_mac": "",
            "device_lan_mac": "",
            "router_lan_ip_address": "",
            "router_wan_ip_address": "",
        }
        gateway.set_gateway_attributes(attributes)
        assert gateway.manufacturer is None
        assert gateway.model is None
        assert gateway.serial_number is None
        assert gateway.firmware_version is None
        assert gateway.hardware_version is None
        assert gateway.primary_connection_type is None
        assert gateway.broadband_connection_status is None
        assert gateway.cellular_connection_status is None
        assert gateway.broadband_lan_mac is None
        assert gateway.device_lan_mac is None
        assert gateway.router_lan_ip_address is None
        assert gateway.router_wan_ip_address is None
        assert gateway._current_poll_interval == ADT_DEFAULT_POLL_INTERVAL
        assert gateway._initial_poll_interval == ADT_DEFAULT_POLL_INTERVAL

    # Check that set_gateway_attributes method sets None for None values and throws an
    # exception for mandatory parameters set to None
    def test_set_gateway_attributes_sets_none_for_none_values(self):
        gateway = ADTPulseGateway()
        attributes = {
            "manufacturer": None,
            "_status_text": None,
            "_current_poll_interval": None,
            "_initial_poll_interval": None,
            "model": None,
            "serial_number": None,
            "next_update": None,
            "last_update": None,
            "firmware_version": None,
            "hardware_version": None,
            "primary_connection_type": None,
            "broadband_connection_status": None,
            "cellular_connection_status": None,
            "cellular_connection_signal_strength": None,
            "broadband_lan_ip_address": None,
            "broadband_lan_mac": None,
            "device_lan_ip_address": None,
            "device_lan_mac": None,
            "router_lan_ip_address": None,
            "router_wan_ip_address": None,
        }
        gateway.set_gateway_attributes(attributes)
        assert gateway.manufacturer is None
        assert gateway._initial_poll_interval == ADT_DEFAULT_POLL_INTERVAL
        assert gateway.model is None
        assert gateway.serial_number is None
        assert gateway.next_update is None
        assert gateway.last_update is None
        assert gateway.firmware_version is None
        assert gateway.hardware_version is None
        assert gateway.primary_connection_type is None
        assert gateway.broadband_connection_status is None
        assert gateway.cellular_connection_status is None
        assert gateway.cellular_connection_signal_strength is None
        assert gateway.broadband_lan_ip_address is None
        assert gateway.broadband_lan_mac is None
        assert gateway.device_lan_ip_address is None
        assert gateway.device_lan_mac is None
        assert gateway.router_lan_ip_address is None
        assert gateway.router_wan_ip_address is None



"""ADT Pulse connection. End users should probably not call this directly."""

import logging
import asyncio
import datetime
import re
import time
from http import HTTPStatus
from random import uniform
from threading import RLock

from aiohttp import (
    ClientConnectionError,
    ClientConnectorError,
    ClientResponse,
    ClientResponseError,
    ClientSession,
)
from bs4 import BeautifulSoup
from yarl import URL

from .const import (
    ADT_DEFAULT_HTTP_ACCEPT_HEADERS,
    ADT_DEFAULT_HTTP_USER_AGENT,
    ADT_DEFAULT_SEC_FETCH_HEADERS,
    ADT_DEFAULT_VERSION,
    ADT_HTTP_BACKGROUND_URIS,
    ADT_LOGIN_URI,
    ADT_LOGOUT_URI,
    ADT_ORB_URI,
    ADT_OTHER_HTTP_ACCEPT_HEADERS,
    ADT_SUMMARY_URI,
    API_HOST_CA,
    API_PREFIX,
    DEFAULT_API_HOST,
    ConnectionFailureReason,
)
from .util import DebugRLock, handle_response, make_soup

RECOVERABLE_ERRORS = {
    HTTPStatus.INTERNAL_SERVER_ERROR,
    HTTPStatus.BAD_GATEWAY,
    HTTPStatus.GATEWAY_TIMEOUT,
}
RETRY_LATER_ERRORS = {HTTPStatus.SERVICE_UNAVAILABLE, HTTPStatus.TOO_MANY_REQUESTS}
LOG = logging.getLogger(__name__)

MAX_RETRIES = 3
SESSION_COOKIES = {"X-mobile-browser": "false", "ICLocal": "en_US"}


class ADTPulseConnection:
    """ADT Pulse connection related attributes."""

    _api_version = ADT_DEFAULT_VERSION
    _class_threadlock = RLock()

    __slots__ = (
        "_api_host",
        "_allocated_session",
        "_session",
        "_attribute_lock",
        "_loop",
        "_last_login_time",
        "_retry_after",
        "_authenticated_flag",
        "_detailed_debug_logging",
        "_site_id",
        "_connection_failure_reason",
    )

    @staticmethod
    def check_service_host(service_host: str) -> None:
        """Check if service host is valid."""
        if service_host is None or service_host == "":
            raise ValueError("Service host is mandatory")
        if service_host not in (DEFAULT_API_HOST, API_HOST_CA):
            raise ValueError(
                f"Service host must be one of {DEFAULT_API_HOST}" f" or {API_HOST_CA}"
            )

    @staticmethod
    def check_login_parameters(username: str, password: str, fingerprint: str) -> None:
        """Check if login parameters are valid.

        Raises ValueError if a login parameter is not valid.
        """
        if username is None or username == "":
            raise ValueError("Username is mandatory")
        pattern = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b"
        if not re.match(pattern, username):
            raise ValueError("Username must be an email address")
        if password is None or password == "":
            raise ValueError("Password is mandatory")
        if fingerprint is None or fingerprint == "":
            raise ValueError("Fingerprint is required")

    @staticmethod
    def get_http_status_description(status_code: int) -> str:
        """Get HTTP status description."""
        status = HTTPStatus(status_code)
        return status.description

    def __init__(
        self,
        host: str,
        session: ClientSession | None = None,
        user_agent: str = ADT_DEFAULT_HTTP_USER_AGENT["User-Agent"],
        debug_locks: bool = False,
        detailed_debug_logging: bool = False,
    ):
        """Initialize ADT Pulse connection."""
        self._attribute_lock: RLock | DebugRLock
        if not debug_locks:
            self._attribute_lock = RLock()
        else:
            self._attribute_lock = DebugRLock("ADTPulseConnection._attribute_lock")
        self._allocated_session = False
        self._authenticated_flag = asyncio.Event()
        if session is None:
            self._allocated_session = True
            self._session = ClientSession()
        else:
            self._session = session
        # need to initialize this after the session since we set cookies
        # based on it
        self.service_host = host
        self._session.headers.update({"User-Agent": user_agent})
        self._session.headers.update(ADT_DEFAULT_HTTP_ACCEPT_HEADERS)
        self._session.headers.update(ADT_DEFAULT_SEC_FETCH_HEADERS)
        self._last_login_time: int = 0
        self._loop: asyncio.AbstractEventLoop | None = None
        self._retry_after = int(time.time())
        self._detailed_debug_logging = detailed_debug_logging
        self._site_id = ""
        self._connection_failure_reason = ConnectionFailureReason.NO_FAILURE

    def __del__(self):
        """Destructor for ADTPulseConnection."""
        if (
            self._allocated_session
            and self._session is not None
            and not self._session.closed
        ):
            self._session.detach()

    @property
    def api_version(self) -> str:
        """Get the API version."""
        with self._class_threadlock:
            return self._api_version

    @property
    def service_host(self) -> str:
        """Get the host prefix for connections."""
        with self._attribute_lock:
            return self._api_host

    @service_host.setter
    def service_host(self, host: str) -> None:
        """Set the host prefix for connections."""
        self.check_service_host(host)
        with self._attribute_lock:
            self._session.headers.update({"Host": host})
            self._api_host = host

    @property
    def loop(self) -> asyncio.AbstractEventLoop | None:
        """Get the event loop."""
        with self._attribute_lock:
            return self._loop

    @loop.setter
    def loop(self, loop: asyncio.AbstractEventLoop | None) -> None:
        """Set the event loop."""
        with self._attribute_lock:
            self._loop = loop

    @property
    def last_login_time(self) -> int:
        """Get the last login time."""
        with self._attribute_lock:
            return self._last_login_time

    @property
    def retry_after(self) -> int:
        """Get the number of seconds to wait before retrying HTTP requests."""
        with self._attribute_lock:
            return self._retry_after

    @retry_after.setter
    def retry_after(self, seconds: int) -> None:
        """Set time after which HTTP requests can be retried."""
        if seconds < time.time():
            raise ValueError("retry_after cannot be less than current time")
        with self._attribute_lock:
            self._retry_after = seconds

    @property
    def authenticated_flag(self) -> asyncio.Event:
        """Get the authenticated flag."""
        with self._attribute_lock:
            return self._authenticated_flag

    @property
    def detailed_debug_logging(self) -> bool:
        """Get detailed debug logging."""
        with self._attribute_lock:
            return self._detailed_debug_logging

    @detailed_debug_logging.setter
    def detailed_debug_logging(self, value: bool) -> None:
        """Set detailed debug logging."""
        with self._attribute_lock:
            self._detailed_debug_logging = value

    @property
    def connection_failure_reason(self) -> ConnectionFailureReason:
        """Get the connection failure reason."""
        with self._attribute_lock:
            return self._connection_failure_reason

    def _set_connection_failure_reason(self, reason: ConnectionFailureReason) -> None:
        """Set the connection failure reason."""
        with self._attribute_lock:
            self._connection_failure_reason = reason

    def check_sync(self, message: str) -> asyncio.AbstractEventLoop:
        """Checks if sync login was performed.

        Returns the loop to use for run_coroutine_threadsafe if so.
        Raises RuntimeError with given message if not."""
        with self._attribute_lock:
            if self._loop is None:
                raise RuntimeError(message)
        return self._loop

    def _set_retry_after(self, code: int, retry_after: str) -> None:
        """
        Check the "Retry-After" header in the response and set retry_after property
        based upon it.

        Parameters:
            code (int): The HTTP response code
            retry_after (str): The value of the "Retry-After" header

        Returns:
            None.
        """
        if retry_after.isnumeric():
            retval = int(retry_after)
        else:
            try:
                retval = int(
                    datetime.datetime.strptime(
                        retry_after, "%a, %d %b %G %T %Z"
                    ).timestamp()
                )
            except ValueError:
                return
        LOG.warning(
            "Task %s received Retry-After %s due to %s",
            asyncio.current_task(),
            retval,
            self.get_http_status_description(code),
        )
        self.retry_after = int(time.time()) + retval
        try:
            fail_reason = ConnectionFailureReason(code)
        except ValueError:
            fail_reason = ConnectionFailureReason.UNKNOWN
        self._set_connection_failure_reason(fail_reason)

    async def async_query(
        self,
        uri: str,
        method: str = "GET",
        extra_params: dict[str, str] | None = None,
        extra_headers: dict[str, str] | None = None,
        timeout: int = 1,
        requires_authentication: bool = True,
    ) -> tuple[int, str | None, URL | None]:
        """
        Query ADT Pulse async.

        Args:
            uri (str): URI to query
            method (str, optional): method to use. Defaults to "GET".
            extra_params (Optional[Dict], optional): query/body parameters.
                                                    Defaults to None.
            extra_headers (Optional[Dict], optional): extra HTTP headers.
                                                    Defaults to None.
            timeout (int, optional): timeout in seconds. Defaults to 1.
            requires_authentication (bool, optional): True if authentication is
                                                    required to perform query.
                                                    Defaults to True.
                                                    If true and authenticated flag not
                                                    set, will wait for flag to be set.

        Returns:
            tuple with integer return code, optional response text, and optional URL of
            response
        """

        async def handle_query_response(
            response: ClientResponse | None,
        ) -> tuple[int, str | None, URL | None, str | None]:
            if response is None:
                return 0, None, None, None
            response_text = await response.text()

            return (
                response.status,
                response_text,
                response.url,
                response.headers.get("Retry-After"),
            )

        if method not in ("GET", "POST"):
            raise ValueError("method must be GET or POST")
        current_time = time.time()
        if self.retry_after > current_time:
            LOG.debug(
                "Retry after set, query %s for %s waiting until %s",
                method,
                uri,
                datetime.datetime.fromtimestamp(self.retry_after),
            )
            await asyncio.sleep(self.retry_after - current_time)

        if requires_authentication and not self.authenticated_flag.is_set():
            LOG.info("%s for %s waiting for authenticated flag to be set", method, uri)
            await self._authenticated_flag.wait()
        else:
            with ADTPulseConnection._class_threadlock:
                if ADTPulseConnection._api_version == ADT_DEFAULT_VERSION:
                    await self.async_fetch_version()

        url = self.make_url(uri)
        headers = extra_headers if extra_headers is not None else {}
        if uri in ADT_HTTP_BACKGROUND_URIS:
            headers.setdefault("Accept", ADT_OTHER_HTTP_ACCEPT_HEADERS["Accept"])
        if self.detailed_debug_logging:
            LOG.debug(
                "Attempting %s %s params=%s timeout=%d",
                method,
                url,
                extra_params,
                timeout,
            )
        retry = 0
        return_value: tuple[int, str | None, URL | None, str | None] = (
            HTTPStatus.OK.value,
            None,
            None,
            None,
        )
        while retry < MAX_RETRIES:
            try:
                async with self._session.request(
                    method,
                    url,
                    headers=extra_headers,
                    params=extra_params if method == "GET" else None,
                    data=extra_params if method == "POST" else None,
                    timeout=timeout,
                ) as response:
                    return_value = await handle_query_response(response)
                    retry += 1

                    if return_value[0] in RECOVERABLE_ERRORS:
                        LOG.info(
                            "query returned recoverable error code %s: %s,"
                            "retrying (count = %d)",
                            return_value[0],
                            self.get_http_status_description(return_value[0]),
                            retry,
                        )
                        if retry == MAX_RETRIES:
                            LOG.warning(
                                "Exceeded max retries of %d, giving up", MAX_RETRIES
                            )
                            response.raise_for_status()
                        await asyncio.sleep(2**retry + uniform(0.0, 1.0))
                        continue

                    response.raise_for_status()
                    break
            except (
                TimeoutError,
                ClientConnectionError,
                ClientConnectorError,
                ClientResponseError,
            ) as ex:
                if return_value[0] is not None and return_value[3] is not None:
                    self._set_retry_after(
                        return_value[0],
                        return_value[3],
                    )
                    break
                LOG.debug(
                    "Error %s occurred making %s request to %s",
                    ex.args,
                    method,
                    url,
                    exc_info=True,
                )
                break
        return (return_value[0], return_value[1], return_value[2])

    def query(
        self,
        uri: str,
        method: str = "GET",
        extra_params: dict[str, str] | None = None,
        extra_headers: dict[str, str] | None = None,
        timeout=1,
        requires_authentication: bool = True,
    ) -> tuple[int, str | None, URL | None]:
        """Query ADT Pulse async.

        Args:
            uri (str): URI to query
            method (str, optional): method to use. Defaults to "GET".
            extra_params (Optional[Dict], optional): query/body parameters. Defaults to None.
            extra_headers (Optional[Dict], optional): extra HTTP headers.
                                                    Defaults to None.
            timeout (int, optional): timeout in seconds. Defaults to 1.
            requires_authentication (bool, optional): True if authentication is required
                                                    to perform query. Defaults to True.
                                                    If true and authenticated flag not
                                                    set, will wait for flag to be set.
        Returns:
            tuple with integer return code, optional response text, and optional URL of
            response
        """
        coro = self.async_query(
            uri,
            method,
            extra_params,
            extra_headers,
            timeout=timeout,
            requires_authentication=requires_authentication,
        )
        return asyncio.run_coroutine_threadsafe(
            coro, self.check_sync("Attempting to run sync query from async login")
        ).result()

    async def query_orb(self, level: int, error_message: str) -> BeautifulSoup | None:
        """Query ADT Pulse ORB.

        Args:
            level (int): error level to log on failure
            error_message (str): error message to use on failure

        Returns:
            Optional[BeautifulSoup]: A Beautiful Soup object, or None if failure
        """
        code, response, url = await self.async_query(
            ADT_ORB_URI,
            extra_headers={"Sec-Fetch-Mode": "cors", "Sec-Fetch-Dest": "empty"},
        )

        return make_soup(code, response, url, level, error_message)

    def make_url(self, uri: str) -> str:
        """Create a URL to service host from a URI.

        Args:
            uri (str): the URI to convert

        Returns:
            str: the converted string
        """
        with self._attribute_lock:
            return f"{self._api_host}{API_PREFIX}{ADTPulseConnection._api_version}{uri}"

    async def async_fetch_version(self) -> None:
        """Fetch ADT Pulse version."""
        response_path: str | None = None
        response_code = HTTPStatus.OK.value
        with ADTPulseConnection._class_threadlock:
            if ADTPulseConnection._api_version != ADT_DEFAULT_VERSION:
                return

            signin_url = f"{self.service_host}"
            try:
                async with self._session.get(
                    signin_url,
                ) as response:
                    response_code = response.status
                    # we only need the headers here, don't parse response
                    response.raise_for_status()
                    response_path = response.url.path
            except (ClientResponseError, ClientConnectionError):
                LOG.warning(
                    "Error %i: occurred during API version fetch, defaulting to %s",
                    response_code,
                    ADT_DEFAULT_VERSION,
                )
                return
            if response_path is not None:
                m = re.search("/myhome/(.+)/[a-z]*/", response_path)
                if m is not None:
                    ADTPulseConnection._api_version = m.group(1)
                    LOG.debug(
                        "Discovered ADT Pulse version %s at %s",
                        ADTPulseConnection._api_version,
                        self.service_host,
                    )
                    return

            LOG.warning(
                "Couldn't auto-detect ADT Pulse version, defaulting to %s",
                ADT_DEFAULT_VERSION,
            )

    async def async_do_login_query(
        self, username: str, password: str, fingerprint: str, timeout: int = 30
    ) -> BeautifulSoup | None:
        """
        Performs a login query to the Pulse site.

        Args:
            timeout (int, optional): The timeout value for the query in seconds.
            Defaults to 30.

        Returns:
            soup: Optional[BeautifulSoup]: A BeautifulSoup object containing
            summary.jsp, or None if failure
            soup: Optional[BeautifulSoup]: A BeautifulSoup object containing
            summary.jsp, or None if failure
        Raises:
            ValueError: if login parameters are not correct
        """

        def extract_seconds_from_string(s: str) -> int:
            seconds = 0
            match = re.search(r"Try again in (\d+)", s)
            if match:
                seconds = int(match.group(1))
                if "minutes" in s:
                    seconds *= 60
            return seconds

        def check_response(
            response: tuple[int, str | None, URL | None]
        ) -> BeautifulSoup | None:
            """Check response for errors."""
            if not handle_response(
                response[0],
                response[2],
                logging.ERROR,
                "Error encountered communicating with Pulse site on login",
            ):
                self._set_connection_failure_reason(ConnectionFailureReason.UNKNOWN)
                return None

            soup = make_soup(
                response[0],
                response[1],
                response[2],
                logging.ERROR,
                "Could not log into ADT Pulse site",
            )
            # FIXME: should probably raise exceptions here
            if soup is None:
                return None
            error = soup.find("div", {"id": "warnMsgContents"})
            if error:
                error_text = error.get_text()
                LOG.error("Error logging into pulse: %s", error_text)
                if retry_after := extract_seconds_from_string(error_text) > 0:
                    self.retry_after = int(time.time()) + retry_after
                return None
            if self.make_url(ADT_SUMMARY_URI) != str(response[2]):
                # more specifically:
                # redirect to signin.jsp = username/password error
                # redirect to mfaSignin.jsp = fingerprint error
                # locked out = error == "Sign In unsuccessful. Your account has been
                # locked after multiple sign in attempts.Try again in 30 minutes."
                LOG.error("Authentication error encountered logging into ADT Pulse")
                self._set_connection_failure_reason(
                    ConnectionFailureReason.INVALID_CREDENTIALS
                )
                return None

            error = soup.find("div", "responsiveContainer")
            if error:
                LOG.error(
                    "2FA authentiation required for ADT pulse username %s: %s",
                    username,
                    error,
                )
                self._set_connection_failure_reason(
                    ConnectionFailureReason.MFA_REQUIRED
                )
                return None
            return soup

        data = {
            "usernameForm": username,
            "passwordForm": password,
            "networkid": self._site_id,
            "fingerprint": fingerprint,
        }
        self.check_login_parameters(username, password, fingerprint)
        try:
            response = await self.async_query(
                ADT_LOGIN_URI,
                "POST",
                extra_params=data,
                timeout=timeout,
                requires_authentication=False,
            )
            if not handle_response(
                response[0],
                response[2],
                logging.ERROR,
                "Error encountered during ADT login GET",
            ):
                return None
        except Exception as e:  # pylint: disable=broad-except
            LOG.error("Could not log into Pulse site: %s", e)
            self._set_connection_failure_reason(ConnectionFailureReason.UNKNOWN)
            return None
        soup = check_response(response)
        if soup is None:
            return None
        with self._attribute_lock:
            self._authenticated_flag.set()
            self._last_login_time = int(time.time())
        return soup

    async def async_do_logout_query(self, site_id: str | None) -> None:
        """Performs a logout query to the ADT Pulse site."""
        params = {}
        if site_id is not None:
            self._site_id = site_id
        params.update({"networkid": self._site_id})

        params.update({"partner": "adt"})
        await self.async_query(
            ADT_LOGOUT_URI,
            extra_params=params,
            timeout=10,
            requires_authentication=False,
        )
        self._authenticated_flag.clear()



# Generated by CodiumAI
import logging
import asyncio
import time
from unittest.mock import Mock

import pytest
from aiohttp import ClientSession
from bs4 import BeautifulSoup

from pyadtpulse import PyADTPulse
from pyadtpulse.const import ADT_SUMMARY_URI, DEFAULT_API_HOST


class TestPyADTPulse:
    # The class can be instantiated with a username, password, and fingerprint.
    def test_instantiation_with_username_password_fingerprint(self):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"

        # Act
        pulse = PyADTPulse(username, password, fingerprint, do_login=False)

        # Assert
        assert pulse._username == username
        assert pulse._password == password
        assert pulse._fingerprint == fingerprint

    # The login method successfully logs in the user.
    @pytest.mark.asyncio
    async def test_login_successfully_logs_in_user(self, mocker):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        pulse = PyADTPulse(username, password, fingerprint, do_login=False)

        async def mock_async_login():
            pulse._pulse_connection.authenticated_flag.set()

        mocker.patch.object(pulse, "async_login", side_effect=mock_async_login)

        # Act
        pulse.login()

        # Assert
        assert pulse._pulse_connection.authenticated_flag.is_set()

    # The logout method successfully logs out the user.
    @pytest.mark.asyncio
    async def test_logout_successfully_logs_out_user(self, mocker):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        pulse = PyADTPulse(username, password, fingerprint, do_login=False)
        pulse._pulse_connection.authenticated_flag.set()

        async def mock_async_logout():
            pulse._pulse_connection.authenticated_flag.clear()

        async def mock_async_login():
            pulse._pulse_connection.authenticated_flag.set()

        mocker.patch.object(
            pulse._pulse_connection,
            "async_do_logout_query",
            side_effect=mock_async_logout,
        )
        mocker.patch.object(pulse, "async_login", side_effect=mock_async_login)

        # Act
        await pulse.logout()

        # Assert
        assert not pulse._pulse_connection.authenticated_flag.is_set()

    # The service_host property can be set to a different host.
    def test_service_host_can_be_set_to_different_host(self):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        pulse = PyADTPulse(username, password, fingerprint, do_login=False)
        new_service_host = DEFAULT_API_HOST

        # Act
        pulse.service_host = new_service_host

        # Assert
        assert pulse._pulse_connection.service_host == new_service_host

    # The relogin_interval property can be set to a different interval.
    def test_relogin_interval_can_be_set_to_different_interval(self):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        pulse = PyADTPulse(username, password, fingerprint, do_login=False)
        new_relogin_interval = 120

        # Act
        pulse.relogin_interval = new_relogin_interval

        # Assert
        assert pulse._relogin_interval == new_relogin_interval

    # The updates_exist property returns True when updates are available.
    def test_updates_exist_returns_true_when_updates_available(self):
        # Arrange
        pulse = PyADTPulse(
            "test@example.com", "password123", "fingerprint123", do_login=False
        )
        pulse._updates_exist.set()

        # Act
        result = pulse.updates_exist

        # Assert
        assert result is True

    # The wait_for_update method waits for updates and returns True when updates are available.
    @pytest.mark.asyncio
    async def test_wait_for_update_returns_true_when_updates_available(self, mocker):
        # Arrange
        pulse = PyADTPulse(
            "test@example.com", "password123", "fingerprint123", do_login=False
        )
        pulse._updates_exist.set()

        # Act
        result = await pulse.wait_for_update()

        # Assert
        assert result is True

    # The is_connected property returns True when the user is connected to the ADT Pulse service.
    def test_is_connected_returns_true_when_user_connected(self, mocker):
        # Arrange
        pulse = PyADTPulse(
            "test@example.com", "password123", "fingerprint123", do_login=False
        )
        pulse._pulse_connection.authenticated_flag.set()
        pulse._pulse_connection.retry_after = time.time() - 1

        # Act
        result = pulse.is_connected

        # Assert
        assert result is True

    # The async_update method successfully queries the ADT Pulse service for updates.
    @pytest.mark.asyncio
    async def test_async_update_successfully_queries_pulse_service_for_updates(
        self, mocker
    ):
        # Arrange
        pulse = PyADTPulse(
            "email@example.com", "password", "fingerprint", do_login=False
        )
        mock_query_orb = mocker.patch.object(pulse._pulse_connection, "query_orb")
        mock_soup = mocker.Mock()
        mock_query_orb.return_value = mock_soup

        # Act
        result = await pulse.async_update()

        # Assert
        assert result is True
        mock_query_orb.assert_called_once_with(
            logging.INFO, "Error returned from ADT Pulse service check"
        )
        pulse._update_sites.assert_called_once_with(mock_soup)

    # The update method successfully queries the ADT Pulse service for updates.
    def test_update_successfully_queries_pulse_service_for_updates(self, mocker):
        # Arrange
        pulse = PyADTPulse(
            "email@example.com", "password", "fingerprint", do_login=False
        )
        mock_async_update = mocker.patch.object(pulse, "async_update")
        mock_async_update.return_value = True

        # Act
        result = pulse.update()

        # Assert
        assert result is True
        mock_async_update.assert_called_once()

    # The sites property returns a list of ADTPulseSite objects.
    def test_sites_property_returns_list_of_sites(self, mocker):
        # Arrange
        pulse = PyADTPulse("email@example.com", "password", "fingerprint")
        mock_site = mocker.Mock()
        pulse._site = mock_site

        # Act
        result = pulse.sites

        # Assert
        assert result == [mock_site]

    # The site property returns an ADTPulseSite object.
    def test_site_property_returns_ADTPulseSite_object(self):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        pulse = PyADTPulse(username, password, fingerprint, do_login=False)

        # Mock the site object
        class MockADTPulseSite:
            pass

        pulse._site = MockADTPulseSite()

        # Act
        site = pulse.site

        # Assert
        assert isinstance(site, MockADTPulseSite)

    # The class can be instantiated with optional parameters such as service_host, user_agent, websession, do_login, debug_locks, keepalive_interval, relogin_interval, and detailed_debug_logging.
    def test_class_instantiation_with_optional_parameters(self):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        service_host = DEFAULT_API_HOST
        user_agent = "Test User Agent"
        websession = ClientSession()
        do_login = False
        debug_locks = True
        keepalive_interval = 10
        relogin_interval = 5
        detailed_debug_logging = True

        # Act
        pulse = PyADTPulse(
            username,
            password,
            fingerprint,
            service_host=service_host,
            user_agent=user_agent,
            websession=websession,
            do_login=do_login,
            debug_locks=debug_locks,
            keepalive_interval=keepalive_interval,
            relogin_interval=relogin_interval,
            detailed_debug_logging=detailed_debug_logging,
        )

        # Assert
        assert pulse._username == username
        assert pulse._password == password
        assert pulse._fingerprint == fingerprint
        assert pulse._pulse_connection.service_host == service_host
        assert pulse._pulse_connection.user_agent == user_agent
        assert pulse._pulse_connection.session == websession
        assert pulse._pulse_connection.debug_locks == debug_locks
        assert pulse.keepalive_interval == keepalive_interval
        assert pulse.relogin_interval == relogin_interval
        assert pulse._detailed_debug_logging == detailed_debug_logging

    # The keepalive_interval property can be set to a different interval.
    def test_keepalive_interval_property_can_be_set(self):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        pulse = PyADTPulse(username, password, fingerprint, do_login=False)
        new_interval = 5

        # Act
        pulse.keepalive_interval = new_interval

        # Assert
        assert pulse.keepalive_interval == new_interval

    # The detailed_debug_logging property can be set to True or False.
    def test_detailed_debug_logging_property(self):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        pulse = PyADTPulse(username, password, fingerprint)

        # Act
        pulse.detailed_debug_logging = True

        # Assert
        assert pulse.detailed_debug_logging == True

        # Act
        pulse.detailed_debug_logging = False

        # Assert
        assert pulse.detailed_debug_logging == False

    # The async_login method returns False if authentication fails.
    @pytest.mark.asyncio
    async def test_async_login_authentication_fails(self, mocker):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        pulse = PyADTPulse(username, password, fingerprint)
        response = Mock()
        response.url = pulse._pulse_connection.make_url(ADT_SUMMARY_URI)
        soup = BeautifulSoup(
            "<div id='warnMsgContents'>Invalid username/password</div>", "html.parser"
        )
        mocker.patch.object(pulse._pulse_connection, "async_fetch_version")
        mocker.patch.object(
            pulse._pulse_connection,
            "async_do_login_query",
            return_value=asyncio.Future(),
        )
        mocker.patch.object(
            pulse._pulse_connection, "query_orb", return_value=asyncio.Future()
        )
        pulse._pulse_connection.async_do_login_query.return_value.set_result(response)
        pulse._pulse_connection.query_orb.return_value.set_result(soup)

        # Act
        result = await pulse.async_login()

        # Assert
        assert result == False

    # The async_logout method successfully logs out the user.
    @pytest.mark.asyncio
    async def test_async_logout_successfully_logs_out_user(self, mocker):
        # Arrange
        pulse = PyADTPulse("test@example.com", "password123", "fingerprint123")
        pulse._pulse_connection.authenticated_flag.set()

        async def mock_async_do_logout_query(site_id):
            return None

        mocker.patch.object(
            pulse._pulse_connection,
            "async_do_logout_query",
            side_effect=mock_async_do_logout_query,
        )

        # Act
        await pulse.async_logout()

        # Assert
        assert not pulse._pulse_connection.authenticated_flag.is_set()

    # The update method returns False if querying the ADT Pulse service fails.
    @pytest.mark.asyncio
    async def test_update_method_returns_false_if_querying_fails(self, mocker):
        # Arrange
        pulse = PyADTPulse("username", "password", "fingerprint")
        mocker.patch.object(pulse._pulse_connection, "query_orb", return_value=None)

        # Act
        result = await pulse.async_update()

        # Assert
        assert result == False

    # The async_update method returns False if querying the ADT Pulse service fails.
    @pytest.mark.asyncio
    async def test_async_update_returns_false_on_query_failure(self, mocker):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        pulse = PyADTPulse(username, password, fingerprint)
        mocker.patch.object(pulse._pulse_connection, "query_orb", return_value=None)

        # Act
        result = await pulse.async_update()

        # Assert
        assert result == False

    # The sites property raises a RuntimeError if no sites have been retrieved.
    def test_sites_property_raises_runtime_error_if_no_sites_retrieved(self, mocker):
        # Arrange
        username = "test@example.com"
        password = "password123"
        fingerprint = "fingerprint123"
        pulse = PyADTPulse(username, password, fingerprint)

        # Act and Assert
        with pytest.raises(RuntimeError):
            pulse.sites



# Generated by CodiumAI
import asyncio
import time

import pytest
from aiohttp import ClientConnectionError, ClientResponseError

from pyadtpulse.const import (
    ADT_DEFAULT_VERSION,
    ADT_LOGIN_URI,
    ADT_LOGOUT_URI,
    ADT_ORB_URI,
    API_HOST_CA,
    API_PREFIX,
    DEFAULT_API_HOST,
)
from pyadtpulse.pulse_connection import ADTPulseConnection


class TestADTPulseConnection:
    # can initialize ADTPulseConnection with valid service host and user agent
    def test_initialize_with_valid_service_host_and_user_agent(self):
        host = DEFAULT_API_HOST
        user_agent = "Test User Agent"
        connection = ADTPulseConnection(host, user_agent=user_agent)

        assert connection.service_host == host
        assert connection._session.headers["User-Agent"] == user_agent

    # can set and get service host
    @pytest.mark.asyncio
    async def test_set_and_get_service_host(self):
        host = DEFAULT_API_HOST
        connection = ADTPulseConnection(host)

        new_host = API_HOST_CA
        connection.service_host = new_host

        assert connection.service_host == new_host

    # can set and get event loop
    @pytest.mark.asyncio
    async def test_set_and_get_event_loop(self):
        host = DEFAULT_API_HOST
        connection = ADTPulseConnection(host)

        loop = asyncio.get_event_loop()
        connection.loop = loop

        assert connection.loop == loop

    # can get last login time
    @pytest.mark.asyncio
    async def test_get_last_login_time(self):
        host = DEFAULT_API_HOST
        connection = ADTPulseConnection(host)

        assert connection.last_login_time == 0

    # can set and get retry after time
    @pytest.mark.asyncio
    async def test_set_and_get_retry_after_time(self):
        connection = ADTPulseConnection(DEFAULT_API_HOST)

        retry_after = int(time.time()) + 60
        connection.retry_after = retry_after

        assert connection.retry_after == retry_after

    # can get authenticated flag
    @pytest.mark.asyncio
    async def test_get_authenticated_flag(self):
        host = DEFAULT_API_HOST
        connection = ADTPulseConnection(host)

        assert not connection.authenticated_flag.is_set()

    # raises ValueError if service host is None or empty string
    @pytest.mark.asyncio
    async def test_raises_value_error_if_service_host_is_none_or_empty_string(self):
        with pytest.raises(ValueError):
            ADTPulseConnection(None)

        with pytest.raises(ValueError):
            ADTPulseConnection("")

    # raises ValueError if service host is not DEFAULT_API_HOST or API_HOST_CA
    @pytest.mark.asyncio
    async def test_raises_value_error_if_service_host_is_not_default_api_host_or_api_host_ca(
        self,
    ):
        with pytest.raises(ValueError):
            ADTPulseConnection("example.com")

        with pytest.raises(ValueError):
            ADTPulseConnection("api.example.com")

    # raises ValueError if retry_after is less than current time
    @pytest.mark.asyncio
    async def test_raises_value_error_if_retry_after_is_less_than_current_time(self):
        host = DEFAULT_API_HOST
        connection = ADTPulseConnection(host)

        with pytest.raises(ValueError):
            connection.retry_after = int(time.time()) - 60

    # can make url with given uri using a valid host
    @pytest.mark.asyncio
    async def test_make_url_with_given_uri_with_valid_host(self):
        host = DEFAULT_API_HOST
        connection = ADTPulseConnection(host)
        uri = "/api/v1/status"
        expected_url = f"{host}/myhome/{ADTPulseConnection._api_version}{uri}"

        url = connection.make_url(uri)

        assert url == expected_url

    # can set and get detailed debug logging
    @pytest.mark.asyncio
    async def test_set_and_get_detailed_debug_logging(self):
        connection = ADTPulseConnection(DEFAULT_API_HOST)

        assert connection.detailed_debug_logging == False

        connection.detailed_debug_logging = True
        assert connection.detailed_debug_logging == True

        connection.detailed_debug_logging = False
        assert connection.detailed_debug_logging == False

    # can do logout query with valid site_id
    @pytest.mark.asyncio
    async def test_logout_query_with_valid_site_id_fixed(self, mocker):
        # Mock the async_query method
        mocker.patch.object(ADTPulseConnection, "async_query")
        connection = ADTPulseConnection(DEFAULT_API_HOST)

        # Set up mock response
        response = mocker.Mock()
        response.status = 200
        ADTPulseConnection.async_query.return_value = response

        # Call the logout query method
        await connection.async_do_logout_query("site_id")

        # Assert that the async_query method was called with the correct parameters
        ADTPulseConnection.async_query.assert_called_once_with(
            ADT_LOGOUT_URI,
            extra_params={"network": "site_id", "partner": "adt"},
            timeout=10,
            requires_authentication=False,
        )

        # Assert that the authenticated_flag was cleared
        assert not connection.authenticated_flag.is_set()

    # can fetch api version with the recommended fix
    @pytest.mark.asyncio
    async def test_fetch_api_version_with_fix(self, mocker):
        # Mock the session and response objects
        session_mock = mocker.Mock()
        response_mock = mocker.Mock()
        response_mock.status = 200
        response_mock.real_url.path = "/myhome/1.0.0/"
        # Mock the session request method to return the mocked response
        session_mock.request.return_value = response_mock

        # Create an instance of ADTPulseConnection with the mocked session
        connection = ADTPulseConnection(DEFAULT_API_HOST, session=session_mock)

        # Call the async_fetch_version method
        await connection.async_fetch_version()

        # Assert that the api_version is updated correctly
        assert connection.api_version == "1.0.0"
        assert connection.service_host == DEFAULT_API_HOST
        assert connection._session.headers["Host"] == f"{DEFAULT_API_HOST}/myhome/1.0.0"

        # Assert that the session request method was called with the correct arguments
        session_mock.request.assert_called_once_with(
            "GET",
            f"{DEFAULT_API_HOST}/myhome/1.0.0",
            headers={"Accept": "application/json"},
            params=None,
            data=None,
            timeout=1,
        )

    # can do login query with valid credentials with fixed host
    @pytest.mark.asyncio
    async def test_login_query_with_valid_credentials_with_fixed_host(self, mocker):
        # Mock the necessary dependencies
        session_mock = mocker.AsyncMock()
        response_mock = mocker.AsyncMock()
        response_mock.status = 200
        response_mock.headers = {"Retry-After": "10"}
        response_mock.real_url.path = "/myhome/1.0/"
        session_mock.request.return_value.__aenter__.return_value = response_mock

        # Create an instance of ADTPulseConnection with fixed host
        connection = ADTPulseConnection(DEFAULT_API_HOST, session=session_mock)

        # Set the authenticated flag to True
        connection._authenticated_flag.set()

        # Mock the make_soup function
        make_soup_mock = mocker.patch("pyadtpulse.util.make_soup")
        make_soup_mock.return_value = None

        # Call the async_query method with valid credentials
        result = await connection.async_query(
            ADT_LOGIN_URI,
            method="POST",
            extra_params={
                "partner": "adt",
                "e": "ns",
                "usernameForm": "test_user",
                "passwordForm": "test_password",
                "fingerprint": "test_fingerprint",
                "sun": "yes",
            },
            timeout=1,
            requires_authentication=False,
        )

        # Assert that the session request method was called with the correct parameters
        session_mock.request.assert_called_once_with(
            "POST",
            f"{DEFAULT_API_HOST}/myhome/login.jsp",
            headers=None,
            params={
                "partner": "adt",
                "e": "ns",
                "usernameForm": "test_user",
                "passwordForm": "test_password",
                "fingerprint": "test_fingerprint",
                "sun": "yes",
            },
            data={
                "partner": "adt",
                "e": "ns",
                "usernameForm": "test_user",
                "passwordForm": "test_password",
                "fingerprint": "test_fingerprint",
                "sun": "yes",
            },
            timeout=1,
        )

        # Assert that the response status is 200
        assert result.status == 200

        # Assert that the make_soup function was called with the correct parameters
        make_soup_mock.assert_called_once_with(response_mock, mocker.ANY, mocker.ANY)

    # raises ClientConnectionError if async_fetch_version fails
    @pytest.mark.asyncio
    async def test_async_fetch_version_raises_error_fixed_fixed(self, mocker):
        # Mock the session object
        session_mock = mocker.Mock()
        session_mock.get.side_effect = ClientConnectionError()

        # Create an instance of ADTPulseConnection with the mocked session
        connection = ADTPulseConnection(DEFAULT_API_HOST, session=session_mock)

        # Call async_fetch_version
        await connection.async_fetch_version()

        # Assert that the api_version remains at its default value
        assert connection.api_version == ADT_DEFAULT_VERSION

    # Raises RuntimeError if async_query is called from sync context with a valid service host
    @pytest.mark.asyncio
    async def test_raises_runtime_error_if_async_query_called_from_sync_context_with_valid_service_host(
        self, mocker
    ):
        # Mock the check_sync method to return a loop
        mocker.patch.object(
            ADTPulseConnection, "check_sync", return_value=asyncio.get_event_loop()
        )

        # Create an instance of ADTPulseConnection with a valid service host
        connection = ADTPulseConnection(DEFAULT_API_HOST)

        # Call the async_query method from a sync context and assert that it raises a RuntimeError
        with pytest.raises(RuntimeError):
            connection.query("uri")

    # async_do_logout_query does not raise an error with a valid service host
    @pytest.mark.asyncio
    async def test_async_do_logout_query_with_valid_service_host(self, mocker):
        # Mock the async_query method to return None
        mocker.patch.object(ADTPulseConnection, "async_query", return_value=None)

        # Create an instance of ADTPulseConnection with a valid service host
        connection = ADTPulseConnection(DEFAULT_API_HOST)

        # Call async_do_logout_query and assert that it does not raise an error
        await connection.async_do_logout_query("site_id")

    # raises RuntimeError if loop is None with a valid service host
    @pytest.mark.asyncio
    async def test_raises_runtime_error_if_loop_is_none_with_valid_service_host(self):
        connection = ADTPulseConnection(DEFAULT_API_HOST)
        connection.loop = None

        with pytest.raises(RuntimeError):
            connection.check_sync("Test message")

    # can set and get allocated session
    @pytest.mark.asyncio
    async def test_set_and_get_allocated_session(self):
        host = DEFAULT_API_HOST
        connection = ADTPulseConnection(host)

        assert connection._allocated_session == True
        assert connection._session is not None

        session = connection._session
        connection._allocated_session = False

        assert connection._allocated_session == False
        assert connection._session == session

    # raises Exception if async_do_login_query fails
    @pytest.mark.asyncio
    async def test_async_do_login_query_failure(self, mocker):
        # Mock the async_query method to raise an exception
        mocker.patch.object(
            ADTPulseConnection,
            "async_query",
            side_effect=Exception("Async query failed"),
        )

        # Create an instance of ADTPulseConnection
        connection = ADTPulseConnection(DEFAULT_API_HOST)

        # Call async_do_login_query and assert that it returns None
        assert (
            await connection.async_do_login_query("username", "password", "fingerprint")
            is None
        )

    # can make url with given uri
    def test_make_url_with_given_uri(self):
        host = DEFAULT_API_HOST
        connection = ADTPulseConnection(host)

        uri = "/test"
        expected_url = f"{host}{API_PREFIX}{ADTPulseConnection._api_version}{uri}"

        assert connection.make_url(uri) == expected_url

    # can fetch api version with async mock (fixed)
    @pytest.mark.asyncio
    async def test_fetch_api_version_with_async_mock_fixed(self, mocker):
        host = DEFAULT_API_HOST
        connection = ADTPulseConnection(host)

        response_mock = mocker.AsyncMock()
        response_mock.real_url.path = "/myhome/v1/"
        mocker.patch.object(response_mock, "__aenter__", return_value=response_mock)

        session_mock = mocker.Mock()
        session_mock.get.return_value = response_mock
        connection._session = session_mock

        await connection.async_fetch_version()

        assert ADTPulseConnection._api_version == "24.0.0-117"

    # can do login query with valid credentials
    @pytest.mark.asyncio
    async def test_do_login_query_with_valid_credentials(self, mocker):
        host = DEFAULT_API_HOST
        connection = ADTPulseConnection(host)

        response_mock = mocker.MagicMock()
        response_mock.__enter__.return_value = response_mock

        session_mock = mocker.MagicMock()
        session_mock.request.return_value = response_mock
        connection._session = session_mock

        username = "me@example.com"
        password = "test_password"
        fingerprint = "test_fingerprint"

        await connection.async_do_login_query(username, password, fingerprint)

        assert connection.authenticated_flag.is_set()

    # raises RuntimeError if loop is None
    def test_raises_runtime_error_if_loop_is_none(self):
        connection = ADTPulseConnection(DEFAULT_API_HOST)
        connection.loop = None

        with pytest.raises(RuntimeError):
            connection.check_sync("Loop is None")

    # raises ClientResponseError if async_fetch_version fails
    @pytest.mark.asyncio
    async def test_raises_client_response_error_if_async_fetch_version_fails(
        self, mocker
    ):
        connection = ADTPulseConnection(DEFAULT_API_HOST)
        mocker.patch.object(connection, "async_query", side_effect=ClientResponseError)

        with pytest.raises(ClientResponseError):
            await connection.async_fetch_version()

    # raises RuntimeError if async_query is called from sync context
    @pytest.mark.asyncio
    async def test_raises_runtime_error_if_async_query_called_from_sync_context(self):
        connection = ADTPulseConnection(DEFAULT_API_HOST)

        with pytest.raises(RuntimeError):
            connection.query(ADT_ORB_URI)



"""ADT Alarm Panel Dataclass."""

import logging
import re
from asyncio import run_coroutine_threadsafe
from dataclasses import dataclass
from threading import RLock
from time import time

from bs4 import BeautifulSoup

from .const import ADT_ARM_DISARM_URI
from .pulse_connection import ADTPulseConnection
from .util import make_soup

LOG = logging.getLogger(__name__)
ADT_ALARM_AWAY = "away"
ADT_ALARM_HOME = "stay"
ADT_ALARM_OFF = "off"
ADT_ALARM_UNKNOWN = "unknown"
ADT_ALARM_ARMING = "arming"
ADT_ALARM_DISARMING = "disarming"

ADT_ARM_DISARM_TIMEOUT: float = 20


@dataclass(slots=True)
class ADTPulseAlarmPanel:
    """ADT Alarm Panel information."""

    model: str = "Unknown"
    _sat: str = ""
    _status: str = "Unknown"
    manufacturer: str = "ADT"
    online: bool = True
    _is_force_armed: bool = False
    _state_lock = RLock()
    _last_arm_disarm: int = int(time())

    @property
    def status(self) -> str:
        """Get alarm status.

        Returns:
            str: the alarm status
        """
        with self._state_lock:
            return self._status

    @property
    def is_away(self) -> bool:
        """Return wheter the system is armed away.

        Returns:
            bool: True if armed away
        """
        with self._state_lock:
            return self._status == ADT_ALARM_AWAY

    @property
    def is_home(self) -> bool:
        """Return whether system is armed at home/stay.

        Returns:
            bool: True if system is armed home/stay
        """
        with self._state_lock:
            return self._status == ADT_ALARM_HOME

    @property
    def is_disarmed(self) -> bool:
        """Return whether the system is disarmed.

        Returns:
            bool: True if the system is disarmed
        """
        with self._state_lock:
            return self._status == ADT_ALARM_OFF

    @property
    def is_force_armed(self) -> bool:
        """Return whether the system is armed in bypass mode.

        Returns:
            bool: True if system armed in bypass mode
        """
        with self._state_lock:
            return self._is_force_armed

    @property
    def is_arming(self) -> bool:
        """Return if system is attempting to arm.

        Returns:
            bool: True if system is attempting to arm
        """
        with self._state_lock:
            return self._status == ADT_ALARM_ARMING

    @property
    def is_disarming(self) -> bool:
        """Return if system is attempting to disarm.

        Returns:
            bool: True if system is attempting to disarm
        """
        with self._state_lock:
            return self._status == ADT_ALARM_DISARMING

    @property
    def last_update(self) -> float:
        """Return last update time.

        Returns:
            float: last arm/disarm time
        """
        with self._state_lock:
            return self._last_arm_disarm

    async def _arm(
        self, connection: ADTPulseConnection, mode: str, force_arm: bool
    ) -> bool:
        """Set arm status.

        Args:
            mode (str)
            force_arm (bool): True if arm force

        Returns:
            bool: True if operation successful
        """
        LOG.debug("Setting ADT alarm %s to %s, force = %s", self._sat, mode, force_arm)
        with self._state_lock:
            if self._status == mode:
                LOG.warning(
                    "Attempting to set alarm status %s to existing status %s",
                    mode,
                    self._status,
                )
            if self._status != ADT_ALARM_OFF and mode != ADT_ALARM_OFF:
                LOG.warning("Cannot set alarm status from %s to %s", self._status, mode)
                return False
            params = {
                "href": "rest/adt/ui/client/security/setArmState",
                "armstate": self._status,  # existing state
                "arm": mode,  # new state
                "sat": self._sat,
            }
            if force_arm and mode != ADT_ALARM_OFF:
                params = {
                    "href": "rest/adt/ui/client/security/setForceArm",
                    "armstate": "forcearm",  # existing state
                    "arm": mode,  # new state
                    "sat": self._sat,
                }

            response = await connection.async_query(
                ADT_ARM_DISARM_URI,
                method="POST",
                extra_params=params,
                timeout=10,
            )

            soup = make_soup(
                response[0],
                response[1],
                response[2],
                logging.WARNING,
                f"Failed updating ADT Pulse alarm {self._sat} to {mode}",
            )
            if soup is None:
                return False

            arm_result = soup.find("div", {"class": "p_armDisarmWrapper"})
            if arm_result is not None:
                error_block = arm_result.find("div")
                if error_block is not None:
                    error_text = arm_result.get_text().replace(
                        "Arm AnywayCancel\n\n", ""
                    )
                    LOG.warning(
                        "Could not set alarm state to %s because %s", mode, error_text
                    )
                    return False
        self._is_force_armed = force_arm
        if mode == ADT_ALARM_OFF:
            self._status = ADT_ALARM_DISARMING
        else:
            self._status = ADT_ALARM_ARMING
        self._last_arm_disarm = int(time())
        return True

    def _sync_set_alarm_mode(
        self,
        connection: ADTPulseConnection,
        mode: str,
        force_arm: bool = False,
    ) -> bool:
        coro = self._arm(connection, mode, force_arm)
        return run_coroutine_threadsafe(
            coro,
            connection.check_sync(
                "Attempting to sync change alarm mode from async session"
            ),
        ).result()

    def arm_away(self, connection: ADTPulseConnection, force_arm: bool = False) -> bool:
        """Arm the alarm in Away mode.

        Args:
            force_arm (bool, Optional): force system to arm

        Returns:
            bool: True if arm succeeded
        """
        return self._sync_set_alarm_mode(connection, ADT_ALARM_AWAY, force_arm)

    def arm_home(self, connection: ADTPulseConnection, force_arm: bool = False) -> bool:
        """Arm the alarm in Home mode.

        Args:
            force_arm (bool, Optional): force system to arm

        Returns:
            bool: True if arm succeeded
        """
        return self._sync_set_alarm_mode(connection, ADT_ALARM_HOME, force_arm)

    def disarm(self, connection: ADTPulseConnection) -> bool:
        """Disarm the alarm.

        Returns:
            bool: True if disarm succeeded
        """
        return self._sync_set_alarm_mode(connection, ADT_ALARM_OFF, False)

    async def async_arm_away(
        self, connection: ADTPulseConnection, force_arm: bool = False
    ) -> bool:
        """Arm alarm away async.

        Args:
            force_arm (bool, Optional): force system to arm

        Returns:
            bool: True if arm succeeded
        """
        return await self._arm(connection, ADT_ALARM_AWAY, force_arm)

    async def async_arm_home(
        self, connection: ADTPulseConnection, force_arm: bool = False
    ) -> bool:
        """Arm alarm home async.

        Args:
            force_arm (bool, Optional): force system to arm
        Returns:
            bool: True if arm succeeded
        """
        return await self._arm(connection, ADT_ALARM_HOME, force_arm)

    async def async_disarm(self, connection: ADTPulseConnection) -> bool:
        """Disarm alarm async.

        Returns:
            bool: True if disarm succeeded
        """
        return await self._arm(connection, ADT_ALARM_OFF, False)

    def update_alarm_from_soup(self, summary_html_soup: BeautifulSoup) -> None:
        """
        Updates the alarm status based on the information extracted from the provided
        HTML soup.

        Args:
            summary_html_soup (BeautifulSoup): The BeautifulSoup object representing
            the HTML soup.

        Returns:
            None: This function does not return anything.
        """
        LOG.debug("Updating alarm status")
        value = summary_html_soup.find("span", {"class": "p_boldNormalTextLarge"})
        sat_location = "security_button_0"
        with self._state_lock:
            if value:
                text = value.text
                last_updated = int(time())

                if re.match("Disarmed", text):
                    if (
                        self._status != ADT_ALARM_ARMING
                        or last_updated - self._last_arm_disarm > ADT_ARM_DISARM_TIMEOUT
                    ):
                        self._status = ADT_ALARM_OFF
                        self._last_arm_disarm = last_updated
                elif re.match("Armed Away", text):
                    if (
                        self._status != ADT_ALARM_DISARMING
                        or last_updated - self._last_arm_disarm > ADT_ARM_DISARM_TIMEOUT
                    ):
                        self._status = ADT_ALARM_AWAY
                        self._last_arm_disarm = last_updated
                elif re.match("Armed Stay", text):
                    if (
                        self._status != ADT_ALARM_DISARMING
                        or last_updated - self._last_arm_disarm > ADT_ARM_DISARM_TIMEOUT
                    ):
                        self._status = ADT_ALARM_HOME
                        self._last_arm_disarm = last_updated
                else:
                    LOG.warning("Failed to get alarm status from '%s'", text)
                    self._status = ADT_ALARM_UNKNOWN
                    self._last_arm_disarm = last_updated
                    return
                LOG.debug("Alarm status = %s", self._status)

            if self._sat == "":
                sat_button = summary_html_soup.find(
                    "input", {"type": "button", "id": sat_location}
                )
                if sat_button and sat_button.has_attr("onclick"):
                    on_click = sat_button["onclick"]
                    match = re.search(r"sat=([a-z0-9\-]+)", on_click)
                    if match:
                        self._sat = match.group(1)
                elif len(self._sat) == 0:
                    LOG.warning("No sat recorded and was unable extract sat.")

                if len(self._sat) > 0:
                    LOG.debug("Extracted sat = %s", self._sat)
                else:
                    LOG.warning("Unable to extract sat")

    def set_alarm_attributes(self, alarm_attributes: dict[str, str]) -> None:
        """
        Set alarm attributes including model, manufacturer, and online status.

        Args:
            self (object): The instance of the alarm.
            alarm_attributes (dict[str, str]): A dictionary containing alarm attributes.

        Returns:
            None
        """
        self.model = alarm_attributes.get("type_model", "Unknown")
        self.manufacturer = alarm_attributes.get("manufacturer_provider", "ADT")
        self.online = alarm_attributes.get("status", "Offline") == "Online"
        LOG.debug(
            "Set alarm attributes: Model = %s, Manufacturer = %s, Online = %s",
            self.model,
            self.manufacturer,
            self.online,
        )
